package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonNull
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import kotlinx.coroutines.runBlocking
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.defs.actor.NPCVariant
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.defs.tile.BuiltinMetaMaterials
import ru.dbotthepony.kstarbound.defs.tile.isNotEmptyLiquid
import ru.dbotthepony.kstarbound.defs.tile.isNotEmptyTile
import ru.dbotthepony.kstarbound.defs.world.TerrestrialWorldParameters
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.nextAABB
import ru.dbotthepony.kstarbound.lua.nextOptionalVector2d
import ru.dbotthepony.kstarbound.lua.nextPoly
import ru.dbotthepony.kstarbound.lua.nextRegistryID
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.nextVector2i
import ru.dbotthepony.kstarbound.lua.nextVector2iOrAABB
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.lua.userdata.LuaFuture
import ru.dbotthepony.kstarbound.lua.userdata.LuaPathFinder
import ru.dbotthepony.kstarbound.lua.userdata.push
import ru.dbotthepony.kstarbound.math.Line2d
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.math.vector.Vector2i
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.util.CarriedExecutor
import ru.dbotthepony.kstarbound.util.GameTimer
import ru.dbotthepony.kstarbound.util.coalesceNull
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.RayCastResult
import ru.dbotthepony.kstarbound.world.RayFilterResult
import ru.dbotthepony.kstarbound.world.TileModification
import ru.dbotthepony.kstarbound.world.TileRayFilter
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.api.AbstractLiquidState
import ru.dbotthepony.kstarbound.world.castRay
import ru.dbotthepony.kstarbound.world.entities.ItemDropEntity
import ru.dbotthepony.kstarbound.world.entities.MonsterEntity
import ru.dbotthepony.kstarbound.world.entities.NPCEntity
import ru.dbotthepony.kstarbound.world.entities.PathFinder
import ru.dbotthepony.kstarbound.world.entities.StagehandEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.world.physics.Poly
import java.util.*
import kotlin.math.PI

private val directionalAngles = intArrayOf(4, 8, 12, 0, 2, 6, 10, 14, 1, 3, 7, 5, 15, 13, 9, 11).let { arr ->
	Array(arr.size) { Vector2d.angle(arr[it] * PI / 8.0) }
}

private fun LuaThread.returnLiquid(liquid: AbstractLiquidState, returnNames: Boolean): Int {
	if (returnNames) {
		pushTable(2)
		setTableValue(1, liquid.state.key)
		setTableValue(2, liquid.level)
		return 1
	} else if (liquid.state.id != null) {
		pushTable(2)
		setTableValue(1, liquid.state.id)
		setTableValue(2, liquid.level)
		return 1
	} else {
		return 0
	}
}

private val LOGGER = LogManager.getLogger()

private fun flyingType(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.sky.flyingType.jsonName)
	return 1
}

private fun warpPhase(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.sky.warpPhase.jsonName)
	return 1
}

private fun skyTime(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.sky.time)
	return 1
}

private fun magnitude(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val arg1 = args.nextVector2d()
	val arg2 = args.nextOptionalVector2d()

	if (arg2 == null) {
		args.lua.push(arg1.length)
	} else {
		args.lua.push(self.geometry.diff(arg1, arg2).length)
	}

	return 1
}

private fun distance(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val arg1 = args.nextVector2d()
	val arg2 = args.nextVector2d()

	args.lua.push(self.geometry.diff(arg1, arg2))

	return 1
}

private fun polyContains(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.geometry.polyContains(args.nextPoly(), args.nextVector2d()))
	return 1
}

private fun xwrap(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val peek = args.peek()

	if (peek == LuaType.NUMBER)
		args.lua.push(self.geometry.x.cell(args.nextDouble()))
	else
		args.lua.push(self.geometry.wrap(args.nextVector2d()))

	return 1
}

private fun ywrap(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val peek = args.peek()

	if (peek == LuaType.NUMBER)
		args.lua.push(self.geometry.y.cell(args.nextDouble()))
	else
		args.lua.push(self.geometry.wrap(args.nextVector2d()))

	return 1
}

private fun nearestTo(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val p0 = args.peek()
	val p1 = args.peek(args.position + 1)

	require(p0 == p1) { "both arguments to world.nearestTo must be of same type (given: $p0 and $p1)" }

	if (p0 == LuaType.NUMBER) {
		args.lua.push(self.geometry.x.nearestTo(args.nextDouble(), args.nextDouble()))
	} else {
		args.lua.push(self.geometry.nearestTo(args.nextVector2d(), args.nextVector2d()))
	}

	return 1
}

private fun readCollisionTypes(args: LuaThread.ArgStack): Set<CollisionType> {
	return when (val type = args.peekAndSkipNothing()) {
		LuaType.TABLE -> EnumSet.copyOf(args.readTableValues { CollisionType.entries.valueOf(args.lua.getString(it) ?: throw IllegalArgumentException("collision list contains non-string values")) })
		LuaType.NIL, LuaType.NONE -> CollisionType.SOLID
		else -> throw IllegalArgumentException("bad argument #${args.position}: table expected, got $type")
	}
}

private fun rectCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val rect = args.nextAABB()
	val collisions = readCollisionTypes(args)
	args.lua.push(self.chunkMap.collide(rect) { it.type in collisions }.findAny().isPresent)

	return 1
}

private fun pointCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val point = args.nextVector2d()
	val collisions = readCollisionTypes(args)
	args.lua.push(self.chunkMap.collide(point) { it.type in collisions })

	return 1
}

private fun pointTileCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val cell = self.getCell(args.nextVector2i())
	val collisions = readCollisionTypes(args)
	args.lua.push(cell.foreground.material.value.collisionKind in collisions)

	return 1
}

private fun castRay(self: World<*, *>, line: Line2d, collisions: Set<CollisionType>): RayCastResult {
	return self.castRay(line, TileRayFilter { cell, fraction, x, y, normal, borderX, borderY -> if (cell.foreground.material.value.collisionKind in collisions) RayFilterResult.HIT else RayFilterResult.SKIP })
}

private fun lineTileCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val line = Line2d(args.nextVector2d(), args.nextVector2d())
	val collisions = readCollisionTypes(args)

	args.lua.push(castRay(self, line, collisions).traversedTiles.isNotEmpty())
	return 1
}

private fun lineTileCollisionPoint(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val line = Line2d(args.nextVector2d(), args.nextVector2d())
	val collisions = readCollisionTypes(args)
	val result = castRay(self, line, collisions)

	if (result.hitTile == null) {
		return 0
	} else {
		args.lua.pushTable(2)
		args.lua.push(1)
		args.lua.push(result.hitTile.borderCross)
		args.lua.setTableValue()
		args.lua.push(2)
		args.lua.push(result.hitTile.normal.normal)
		args.lua.setTableValue()
		return 1
	}
}

private fun rectTileCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val rect = args.nextAABB()
	val collisions = readCollisionTypes(args)
	args.lua.push(self.chunkMap.anyCellSatisfies(rect, World.CellPredicate { x, y, cell -> cell.foreground.material.value.collisionKind in collisions }))

	return 1
}

private fun lineCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val line = Line2d(args.nextVector2d(), args.nextVector2d())
	val collisions = readCollisionTypes(args)
	val result = self.chunkMap.collide(line) { it.type in collisions }

	if (result == null) {
		return 0
	} else {
		args.lua.push(result.border)
		args.lua.push(result.normal)
		return 2
	}
}

private fun polyCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val poly = args.nextPoly()
	val translation = args.nextOptionalVector2d()
	val collisions = readCollisionTypes(args)
	val result = self.chunkMap.collide(if (translation == null) poly else poly + translation) { it.type in collisions }

	args.lua.push(result.findAny().isPresent)
	return 1
}

private fun collisionBlocksAlongLine(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val line = Line2d(args.nextVector2d(), args.nextVector2d())
	val collisions = readCollisionTypes(args)
	var limit = args.nextOptionalLong() ?: Long.MAX_VALUE

	if (limit < 0L)
		limit = Long.MAX_VALUE

	val result = self.castRay(line) { cell, fraction, x, y, normal, borderX, borderY ->
		if (cell.foreground.material.value.collisionKind in collisions) {
			val tlimit = --limit

			if (tlimit > 0L) {
				RayFilterResult.CONTINUE
			} else if (tlimit == 0L) {
				RayFilterResult.HIT
			} else {
				RayFilterResult.BREAK
			}
		} else {
			RayFilterResult.SKIP
		}
	}

	args.lua.pushTable(result.traversedTiles.size)

	for ((i, data) in result.traversedTiles.withIndex()) {
		args.lua.push(i + 1L)
		args.lua.push(data.pos)
		args.lua.setTableValue()
	}

	return 1
}

private fun liquidAlongLine(self: World<*, *>, names: Boolean, args: LuaThread.ArgStack): Int {
	var i = 1L

	args.lua.pushTable()

	self.castRay(args.nextVector2d(), args.nextVector2d()) { cell, fraction, x, y, normal, borderX, borderY ->
		if (cell.liquid.state.isNotEmptyLiquid && cell.liquid.level > 0f && (names || cell.liquid.state.id != null)) {
			args.lua.push(i++)
			args.lua.pushTable(2)

			args.lua.push(1)
			args.lua.push(Vector2i(x, y))
			args.lua.setTableValue()

			args.lua.push(2)
			args.lua.pushTable(2)

			if (names)
				args.lua.setTableValue(1, cell.liquid.state.key)
			else
				args.lua.setTableValue(1, cell.liquid.state.id!!)

			args.lua.setTableValue(2, cell.liquid.level.toDouble())
			args.lua.setTableValue()

			args.lua.setTableValue()
		}

		RayFilterResult.SKIP
	}

	return 1
}

private fun resolvePolyCollision(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val originalPoly = args.nextPoly()
	val position = args.nextVector2d()
	val maximumCorrection = args.nextDouble()
	val collisions = readCollisionTypes(args)
	val poly = originalPoly + position

	data class Entry(val poly: Poly, val center: Vector2d, var distance: Double = 0.0) : Comparable<Entry> {
		override fun compareTo(other: Entry): Int {
			return distance.compareTo(other.distance)
		}
	}

	val tiles = ObjectArrayList<Entry>()

	for (tile in self.chunkMap.queryTileCollisions(poly.aabb.enlarge(maximumCorrection + 1.0, maximumCorrection + 1.0))) {
		if (tile.type in collisions) {
			tiles.add(Entry(tile.poly, tile.poly.centre))
		}
	}

	fun separate(loops: Int, axis: Vector2d?): Vector2d? {
		var body = poly
		var correction = Vector2d.ZERO

		for (i in 0 until loops) {
			val center = body.centre

			for (tile in tiles)
				tile.distance = tile.center.distanceSquared(center)

			tiles.sort()

			var anyIntersects = false

			for (tile in tiles) {
				val intersection = tile.poly.intersect(body, axis)

				if (intersection != null) {
					anyIntersects = true
					body += intersection.vector
					correction += intersection.vector

					if (correction.length >= maximumCorrection)
						return null
				}
			}

			if (!anyIntersects)
				return correction
		}

		for (tile in tiles) {
			if (tile.poly.intersect(body) != null) {
				return null
			}
		}

		return correction
	}

	var trySeparate = separate(10, null)

	// First try any-directional SAT separation for two loops
	if (trySeparate != null) {
		args.lua.push(position + trySeparate)
		return 1
	}

	// Then, try direction-limiting SAT in cardinals, then 45 degs, then in
	// between, for 16 total angles in a circle.
	// TODO: if "any direction" separation fails, then this will surely fail too?
	for (axis in directionalAngles) {
		trySeparate = separate(4, axis)

		if (trySeparate != null) {
			args.lua.push(position + trySeparate)
			return 1
		}
	}

	return 0
}

private fun tileIsOccupied(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val cell = self.getCell(pos)
	val isForeground = args.nextOptionalBoolean()
	val includeEphemeral = args.nextOptionalBoolean() == true

	if (cell.tile(isForeground == false).material.isNotEmptyTile) {
		args.lua.push(true)
	} else {
		args.lua.push(self.entityIndex.tileEntitiesAt(pos).any { !it.isEphemeral || includeEphemeral })
	}

	return 1
}

private fun placeObject(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val type = args.nextString()
	val pos = args.nextVector2i()
	val objectDirection = args.nextOptionalLong()?.toInt()
	var parameters = args.nextOptionalJson()

	if (parameters !is JsonObject)
		parameters = JsonObject()

	try {
		val prototype = Registries.worldObjects[type] ?: throw IllegalStateException("No such object $type")
		var direction = Direction.RIGHT

		if (objectDirection != null && objectDirection.toLong() < 0L)
			direction = Direction.LEFT

		val orientation = prototype.value.findValidOrientation(self, pos, direction)

		if (orientation == -1) {
			LOGGER.debug("Lua script tried to place object {} at {}, but it can't be placed there!", prototype.key, pos)
			args.lua.push(false)
		} else {
			val create = WorldObject.create(prototype, pos, parameters)
			create?.orientationIndex = orientation.toLong()
			create?.joinWorld(self)
			args.lua.push(create != null)
		}
	} catch (err: Throwable) {
		LOGGER.error("Exception while placing world object $type at $pos", err)
		args.lua.push(false)
	}

	return 1
}

private fun spawnItem(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val descriptor: ItemDescriptor
	val pos: Vector2d
	val peek = args.peek()

	if (peek == LuaType.TABLE) {
		descriptor = ItemDescriptor(args)
		pos = args.nextVector2d()
		args.skip(2)
	} else if (peek == LuaType.STRING) {
		val name = args.nextString()
		pos = args.nextVector2d()
		val count = args.nextOptionalLong() ?: 1L
		val parameters = args.nextOptionalJson()?.coalesceNull?.asJsonObject ?: JsonObject()
		descriptor = ItemDescriptor(name, count, parameters)
	} else {
		throw IllegalArgumentException("bad argument #1 to spawnItem: ItemDescriptor expected, got $peek")
	}

	val initialVelocity = args.nextOptionalVector2d() ?: Vector2d.ZERO
	val intangibleTime = args.nextOptionalDouble() ?: 0.0

	try {
		if (descriptor.isEmpty) {
			if (LOGGER.isDebugEnabled)
				LOGGER.debug("Lua script tried to create non existing item $descriptor at $pos")

			return 0
		} else {
			val create = ItemDropEntity(descriptor, args.lua.random)
			create.movement.velocity = initialVelocity
			create.intangibleTimer = GameTimer(intangibleTime)

			create.joinWorld(self)
			args.lua.push(create.entityID.toLong())
			return 1
		}
	} catch (err: Throwable) {
		LOGGER.error("Exception while creating item $descriptor at $pos", err)
		return 0
	}
}

private fun spawnTreasure(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2d()
	val pool = args.nextString()
	val level = args.nextDouble()
	val seed = args.nextOptionalLong()

	args.lua.pushTable()
	var i = 1L

	try {
		val items = Registries.treasurePools
			.getOrThrow(pool)
			.value
			.evaluate(if (seed != null) random(seed.toLong()) else args.lua.random, level)

		for (item in items) {
			val entity = ItemDropEntity(item)
			entity.position = pos
			entity.joinWorld(self)

			args.lua.push(i++)
			args.lua.push(entity.entityID.toLong())
			args.lua.setTableValue()
		}
	} catch (err: Throwable) {
		LOGGER.error("Exception while spawning treasure from pool '$pool' at $pos", err)
	}

	return 1
}

private fun spawnMonster(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val type = args.nextString()
	val position = args.nextVector2d()
	val overrides = args.nextOptionalJson()

	try {
		val parameters = JsonObject()
		parameters["aggressive"] = args.lua.random.nextBoolean()

		if (overrides != null) {
			mergeJson(parameters, overrides)
		}

		val level = parameters["level"]?.asDouble ?: 1.0
		val seed: Long

		if ("seed" !in parameters || !parameters["seed"].asJsonPrimitive.isNumber) {
			seed = args.lua.random.nextLong()
		} else {
			seed = parameters["seed"].asLong
		}

		val variant = Registries.monsterTypes.getOrThrow(type).value.create(seed, parameters)
		val monster = MonsterEntity(variant, level)
		monster.position = position
		monster.joinWorld(self)
		args.lua.push(monster.entityID.toLong())
		return 1
	} catch (err: Throwable) {
		LOGGER.error("Exception caught while spawning Monster type $type", err)
	}

	return 0
}

private fun spawnNpc(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val position = args.nextVector2d()
	val species = args.nextString()
	val type = args.nextString()
	val level = args.nextDouble()
	val seed = args.nextOptionalLong() ?: args.lua.random.nextLong()
	val overrides = args.nextOptionalJson() ?: JsonNull.INSTANCE

	try {
		// TODO: this blocks world thread
		val npc = runBlocking {
			NPCEntity(NPCVariant.create(
				Registries.species.getOrThrow(species),
				type,
				level,
				seed,
				overrides
			))
		}

		npc.position = position
		npc.joinWorld(self)
		args.lua.push(npc.entityID.toLong())
		return 1
	} catch (err: Throwable) {
		LOGGER.error("Exception caught while spawning NPC $type with species $species", err)
	}

	return 0
}

private fun spawnStagehand(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val position = args.nextVector2d()
	val type = args.nextString()
	val overrides = args.nextOptionalJson() ?: JsonNull.INSTANCE

	try {
		val stagehand = StagehandEntity(type, overrides)
		stagehand.position = position
		stagehand.joinWorld(self)
		args.lua.push(stagehand.entityID.toLong())
		return 1
	} catch (err: Throwable) {
		LOGGER.error("Exception caught while spawning stagehand of type '$type'", err)
	}

	return 0
}

private fun threatLevel(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.threatLevel)
	return 1
}

private fun day(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.sky.day.toLong())
	return 1
}

private fun timeOfDay(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.sky.timeOfDay)
	return 1
}

private fun dayLength(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.sky.dayLength)
	return 1
}

private fun getProperty(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	args.lua.push(self.getProperty(name) { args.nextOptionalJson() ?: JsonNull.INSTANCE })
	return 1
}

private fun setProperty(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	self.setProperty(name, args.nextOptionalJson() ?: JsonNull.INSTANCE)
	return 1
}

private fun liquidAt(self: World<*, *>, names: Boolean, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2iOrAABB()

	if (pos.isLeft) {
		val cell = self.getCell(pos.left())

		if (cell.liquid.state.isNotEmptyLiquid) {
			return args.lua.returnLiquid(cell.liquid, names)
		}
	} else {
		val level = self.averageLiquidLevel(pos.right())

		if (level != null && (names || level.type.id != null)) {
			args.lua.pushTable(2)

			if (names)
				args.lua.setTableValue(1, level.type.key)
			else
				args.lua.setTableValue(1, level.type.id)

			args.lua.setTableValue(2, level.average)
			return 1
		}
	}

	return 0
}

private fun gravity(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.chunkMap.gravityAt(args.nextVector2d()).y)
	return 1
}

private fun gravityVector(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.chunkMap.gravityAt(args.nextVector2d()))
	return 1
}

private fun spawnLiquid(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val id = args.nextRegistryID()
	val quantity = args.nextDouble()

	val action = TileModification.Pour(Registries.liquid.ref(id), quantity.toFloat())
	args.lua.push(self.applyTileModifications(listOf(pos to action), false).thenApply { it.isEmpty() }.getNow(true))
	return 1
}

private fun spawnLiquidPromise(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val id = args.nextRegistryID()
	val quantity = args.nextDouble()
	val action = TileModification.Pour(Registries.liquid.ref(id), quantity.toFloat())

	args.lua.push(
		LuaFuture(
			future = self.applyTileModifications(listOf(pos to action), false).thenApply { it.isEmpty() },
			isLocal = false,
			handler = {
				push(it)
				1
			}
		)
	)

	return 1
}

private fun destroyLiquid(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val action = TileModification.Pour(BuiltinMetaMaterials.NO_LIQUID.ref, 0f)
	val cell = self.getCell(pos)
	self.applyTileModifications(listOf(pos to action), false)

	if (cell.liquid.state.isNotEmptyLiquid) {
		return args.lua.returnLiquid(cell.liquid, false)
	}

	return 0
}

private fun destroyLiquidPromise(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val action = TileModification.Pour(BuiltinMetaMaterials.NO_LIQUID.ref, 0f)
	val cell = self.getCell(pos).immutable()
	self.applyTileModifications(listOf(pos to action), false)

	args.lua.push(
		LuaFuture(
			self.applyTileModifications(listOf(pos to action), false).thenApply { it.isEmpty() },
			false,
			{
				push(it)
				returnLiquid(cell.liquid, true) + 1
			}
		)
	)

	return 1
}

private fun isTileProtected(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.isDungeonIDProtected(self.getCell(args.nextVector2i()).dungeonId))
	return 1
}

private fun findPlatformerPath(self: World<*, *>, args: LuaThread.ArgStack): Int {
	LOGGER.warn("world.findPlatformerPath() was called, this will cause world lag. Consider switching to world.platformerPathStart(), since it is multithreaded in new engine.")

	val start = args.nextVector2d()
	val end = args.nextVector2d()
	val actorParams = args.nextJson()
	val searchParams = args.nextJson()

	val finder = PathFinder(
		self,
		start,
		end,
		Starbound.gson.fromJsonFast(actorParams, ActorMovementParameters::class.java),
		Starbound.gson.fromJsonFast(searchParams, PathFinder.Parameters::class.java))

	finder.run(Int.MAX_VALUE)
	return LuaPathFinder.convertPath(args.lua, finder.result.orNull())
}

private fun platformerPathStart(self: World<*, *>, pacer: CarriedExecutor, args: LuaThread.ArgStack): Int {
	val start = args.nextVector2d()
	val end = args.nextVector2d()
	val actorParams = args.nextJson()
	val searchParams = args.nextJson()

	args.lua.push(LuaPathFinder(pacer, PathFinder(
		self,
		start,
		end,
		Starbound.gson.fromJsonFast(actorParams, ActorMovementParameters::class.java),
		Starbound.gson.fromJsonFast(searchParams, PathFinder.Parameters::class.java))))

	return 1
}

private fun type(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.worldParameters?.typeName ?: "unknown")
	return 1
}

private fun size(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.geometry.size)
	return 1
}

private fun inSurfaceLayer(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.isSurfaceLayer(args.nextVector2i()))
	return 1
}

private fun surfaceLevel(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.surfaceLevel.toLong())
	return 1
}

private fun terrestrial(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.worldParameters is TerrestrialWorldParameters)
	return 1
}

private fun itemDropItem(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val entity = self.entities[args.nextInt()] as? ItemDropEntity ?: return 0
	entity.item.createDescriptor().store(args.lua)
	return 1
}

private fun biomeBlocksAt(self: World<*, *>, names: Boolean, args: LuaThread.ArgStack): Int {
	val info = self.template.cellInfo(args.nextVector2i())
	args.lua.pushTable()
	var i = 1L

	val biome = info.blockBiome

	if (biome != null) {
		if (names) {
			biome.mainBlock.native.entry?.key?.let { args.lua.setTableValue(i++, it) }
			biome.subBlocks.forEach { it.native.entry?.key?.let { args.lua.setTableValue(i++, it) } }
		} else {
			biome.mainBlock.native.entry?.id?.let { args.lua.setTableValue(i++, it) }
			biome.subBlocks.forEach { it.native.entry?.id?.let { args.lua.setTableValue(i++, it) } }
		}

	}

	return 1
}

private fun dungeonId(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.getCell(args.nextVector2i()).dungeonId.toLong())
	return 1
}

private val worldScript by lazy { LuaThread.loadInternalScript("world") }

fun provideWorldBindings(self: World<*, *>, lua: LuaThread) {
	lua.pushTable()
	lua.dup()
	lua.storeGlobal("world")

	lua.pushBinding(self, "flyingType", ::flyingType)
	lua.pushBinding(self, "warpPhase", ::warpPhase)
	lua.pushBinding(self, "skyTime", ::skyTime)
	lua.pushBinding(self, "time", ::skyTime)

	lua.pushBinding(self, "magnitude", ::magnitude)
	lua.pushBinding(self, "distance", ::distance)

	lua.pushBinding(self, "polyContains", ::polyContains)

	lua.pushBinding(self, "xwrap", ::xwrap)
	lua.pushBinding(self, "ywrap", ::ywrap)
	lua.pushBinding(self, "nearestTo", ::nearestTo)

	lua.pushBinding(self, "rectCollision", ::rectCollision)
	lua.pushBinding(self, "polyCollision", ::polyCollision)
	lua.pushBinding(self, "pointCollision", ::pointCollision)
	lua.pushBinding(self, "pointTileCollision", ::pointTileCollision)
	lua.pushBinding(self, "lineTileCollision", ::lineTileCollision)
	lua.pushBinding(self, "lineTileCollisionPoint", ::lineTileCollisionPoint)
	lua.pushBinding(self, "rectTileCollision", ::rectTileCollision)
	lua.pushBinding(self, "lineCollision", ::lineCollision)

	lua.pushBinding(self, "collisionBlocksAlongLine", ::collisionBlocksAlongLine)
	lua.pushBinding(self, false, "liquidAlongLine", ::liquidAlongLine)
	lua.pushBinding(self, true, "liquidNamesAlongLine", ::liquidAlongLine)

	lua.pushBinding(self, "resolvePolyCollision", ::resolvePolyCollision)
	lua.pushBinding(self, "tileIsOccupied", ::tileIsOccupied)
	lua.pushBinding(self, "placeObject", ::placeObject)
	lua.pushBinding(self, "spawnItem", ::spawnItem)
	lua.pushBinding(self, "spawnTreasure", ::spawnTreasure)
	lua.pushBinding(self, "spawnMonster", ::spawnMonster)
	lua.pushBinding(self, "spawnNpc", ::spawnNpc)
	lua.pushBinding(self, "spawnStagehand", ::spawnStagehand)

	lua.setTableValueToStub("spawnProjectile")
	lua.setTableValueToStub("spawnVehicle")

	lua.pushBinding(self, "threatLevel", ::threatLevel)
	lua.pushBinding(self, "day", ::day)
	lua.pushBinding(self, "timeOfDay", ::timeOfDay)
	lua.pushBinding(self, "dayLength", ::dayLength)

	lua.pushBinding(self, "getProperty", ::getProperty)
	lua.pushBinding(self, "setProperty", ::setProperty)

	lua.pushBinding(self, false, "liquidAt", ::liquidAt)
	lua.pushBinding(self, true, "liquidNameAt", ::liquidAt)

	lua.pushBinding(self, "gravity", ::gravity)
	lua.pushBinding(self, "gravityVector", ::gravityVector)

	lua.pushBinding(self, "spawnLiquid", ::spawnLiquid)
	lua.pushBinding(self, "spawnLiquidPromise", ::spawnLiquidPromise)

	lua.pushBinding(self, "destroyLiquid", ::destroyLiquid)
	lua.pushBinding(self, "destroyLiquidPromise", ::destroyLiquidPromise)

	lua.pushBinding(self, "isTileProtected", ::isTileProtected)

	lua.pushBinding(self, "findPlatformerPath", ::findPlatformerPath)
	lua.pushBinding(self, CarriedExecutor(Starbound.EXECUTOR), "platformerPathStart", ::platformerPathStart)

	lua.pushBinding(self, "type", ::type)
	lua.pushBinding(self, "size", ::size)
	lua.pushBinding(self, "inSurfaceLayer", ::inSurfaceLayer)
	lua.pushBinding(self, "surfaceLevel", ::surfaceLevel)
	lua.pushBinding(self, "terrestrial", ::terrestrial)
	lua.pushBinding(self, "itemDropItem", ::itemDropItem)

	lua.pushBinding(self, false, "biomeBlocksAt", ::biomeBlocksAt)
	lua.pushBinding(self, true, "biomeBlockNamesAt", ::biomeBlocksAt)

	lua.pushBinding(self, "dungeonId", ::dungeonId)

	lua.setTableValueToEmpty("debugPoint")
	lua.setTableValueToEmpty("debugLine")
	lua.setTableValueToEmpty("debugPoly")
	lua.setTableValueToEmpty("debugText")

	provideWorldEnvironmentalBindings(self, lua)
	provideWorldEntitiesBindings(self, lua)

	if (self is ServerWorld) {
		provideServerWorldBindings(self, lua)
	}

	lua.pop()

	lua.load(worldScript, "@/internal/world.lua")
	lua.call()
}
