package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.defs.`object`.LoungeOrientation
import ru.dbotthepony.kstarbound.item.ItemStack
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.nextAABB
import ru.dbotthepony.kstarbound.lua.nextPoly
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.nextVector2i
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.lua.setTableValue
import ru.dbotthepony.kstarbound.lua.userdata.LuaFuture
import ru.dbotthepony.kstarbound.lua.userdata.push
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Line2d
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.util.random.shuffle
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.ActorEntity
import ru.dbotthepony.kstarbound.world.entities.DynamicEntity
import ru.dbotthepony.kstarbound.world.entities.HumanoidActorEntity
import ru.dbotthepony.kstarbound.world.entities.ItemDropEntity
import ru.dbotthepony.kstarbound.world.entities.MonsterEntity
import ru.dbotthepony.kstarbound.world.entities.NPCEntity
import ru.dbotthepony.kstarbound.world.entities.StagehandEntity
import ru.dbotthepony.kstarbound.world.entities.api.InspectableEntity
import ru.dbotthepony.kstarbound.world.entities.api.InteractiveEntity
import ru.dbotthepony.kstarbound.world.entities.api.ScriptedEntity
import ru.dbotthepony.kstarbound.world.entities.player.PlayerEntity
import ru.dbotthepony.kstarbound.world.entities.tile.ContainerObject
import ru.dbotthepony.kstarbound.world.entities.tile.LoungeableObject
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import ru.dbotthepony.kstarbound.world.physics.Poly
import java.util.*
import java.util.concurrent.CompletableFuture
import java.util.function.Predicate

private enum class EntityBoundMode {
	META_BOUNDING_BOX,
	COLLISION_AREA,
	POSITION;
}

private enum class Order {
	WHATEVER, RANDOM, NEAREST
}

private fun LuaThread.ArgStack.getTypes(): Set<EntityType> {
	if (peek().isNothing) {
		return EntityType.ALL
	} else {
		val types = EnumSet.noneOf(EntityType::class.java)

		iterateTable(keyVisitor = {}, valueVisitor = {
			when (val type = getLong(it)!!.toInt()) {
				// mobile
				10 -> {
					types.add(EntityType.PLAYER)
					types.add(EntityType.MONSTER)
					types.add(EntityType.NPC)
					types.add(EntityType.PROJECTILE)
					types.add(EntityType.ITEM_DROP)
					types.add(EntityType.VEHICLE)
				}

				// creature
				11 -> {
					types.add(EntityType.PLAYER)
					types.add(EntityType.MONSTER)
					types.add(EntityType.NPC)
				}

				// specific
				else -> types.add(EntityType.entries[type])
			}
		})

		return types
	}
}

private data class CallScriptData(
	val name: String,
	val arguments: JsonArray,
	val expectedResult: JsonElement
)

private fun LuaThread.ArgStack.getScriptData(): CallScriptData? {
	if (peek() == LuaType.STRING) {
		return CallScriptData(nextString(), nextJson().asJsonArray, nextJson())
	} else {
		skip(3)
		return null
	}
}

private fun AbstractEntity.genericCheck(
	predicate: Predicate<AbstractEntity>,
	scriptData: CallScriptData?,
	withoutEntityId: Int?,
	includeTypes: Set<EntityType>
): Boolean {
	if (!predicate.test(this)) return false
	if (entityID == withoutEntityId) return false
	if (type !in includeTypes) return false

	if (scriptData != null) {
		if (isRemote || this !is ScriptedEntity) return false
		val results = callScript(scriptData.name, scriptData.arguments)
		if (results != scriptData.expectedResult) return false
	}

	return true
}

private data class GenericQueryData(
	val includeTypes: Set<EntityType>,
	val withoutEntityId: Int?,
	val boundMode: EntityBoundMode,
	val scriptData: CallScriptData?,
	val order: Order
)

private fun genericQueryData(args: LuaThread.ArgStack): GenericQueryData {
	val includeTypes = args.getTypes()
	val withoutEntityId = args.nextOptionalLong()?.toInt()
	val boundMode = EntityBoundMode.entries[args.nextInt()]
	val scriptData = args.getScriptData()
	val order = Order.entries[args.nextInt()]

	return GenericQueryData(includeTypes, withoutEntityId, boundMode, scriptData, order)
}

private fun LuaThread.returnQueryResult(results: List<AbstractEntity>): Int {
	pushTable(results.size)

	for ((i, entity) in results.withIndex()) {
		setTableValue(i + 1L, entity.entityID.toLong())
	}

	return 1
}

private fun queryAABBImpl(self: World<*, *>, predicate: Predicate<AbstractEntity>, args: LuaThread.ArgStack): Int {
	val aabb = args.nextAABB()
	val (includeTypes, withoutEntityId, boundMode, scriptData, order) = genericQueryData(args)

	val innerPredicate: Predicate<AbstractEntity> = when (boundMode) {
		EntityBoundMode.META_BOUNDING_BOX -> Predicate p@{
			it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)
		}

		EntityBoundMode.COLLISION_AREA -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			val collisionArea = it.collisionArea
			return@p collisionArea.isEmpty || self.geometry.rectIntersectsRect(aabb, collisionArea)
		}

		EntityBoundMode.POSITION -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			return@p self.geometry.rectContains(aabb, it.position)
		}
	}

	val entities = self.entityIndex.query(aabb, innerPredicate)

	when (order) {
		Order.WHATEVER -> {} // as-is
		Order.RANDOM -> entities.shuffle(args.lua.random)
		Order.NEAREST -> {
			val nearest = aabb.centre
			entities.sortWith { o1, o2 -> self.geometry.diff(o1.position, nearest).lengthSquared.compareTo(self.geometry.diff(o2.position, nearest).lengthSquared) }
		}
	}

	return args.lua.returnQueryResult(entities)
}

private fun queryRadiusImpl(self: World<*, *>, predicate: Predicate<AbstractEntity>, args: LuaThread.ArgStack): Int {
	val point = args.nextVector2d()
	val radius = args.nextDouble()
	val (includeTypes, withoutEntityId, boundMode, scriptData, order) = genericQueryData(args)

	val innerPredicate: Predicate<AbstractEntity> = when (boundMode) {
		EntityBoundMode.META_BOUNDING_BOX -> Predicate p@{
			it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes) && self.geometry.rectIntersectsCircle(it.metaBoundingBox, point, radius)
		}

		EntityBoundMode.COLLISION_AREA -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			var collisionArea = it.collisionArea

			if (collisionArea.isEmpty)
				collisionArea = it.metaBoundingBox

			return@p self.geometry.rectIntersectsCircle(collisionArea, point, radius)
		}

		EntityBoundMode.POSITION -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			return@p self.geometry.diff(point, it.position).length <= radius
		}
	}

	val entities = self.entityIndex.query(AABB.withSide(point, radius), innerPredicate)

	when (order) {
		Order.WHATEVER -> {} // as-is
		Order.RANDOM -> entities.shuffle(args.lua.random)
		Order.NEAREST -> {
			entities.sortWith { o1, o2 -> self.geometry.diff(o1.position, point).lengthSquared.compareTo(self.geometry.diff(o2.position, point).lengthSquared) }
		}
	}

	return args.lua.returnQueryResult(entities)
}

private fun queryLineImpl(self: World<*, *>, predicate: Predicate<AbstractEntity>, args: LuaThread.ArgStack): Int {
	val line = Line2d(args.nextVector2d(), args.nextVector2d())
	val (includeTypes, withoutEntityId, boundMode, scriptData, order) = genericQueryData(args)

	val innerPredicate: Predicate<AbstractEntity> = when (boundMode) {
		EntityBoundMode.META_BOUNDING_BOX -> Predicate p@{
			it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)
		}

		EntityBoundMode.COLLISION_AREA -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			var collisionArea = it.collisionArea

			if (collisionArea.isEmpty)
				collisionArea = it.metaBoundingBox

			return@p self.geometry.lineIntersectsRect(line, collisionArea)
		}

		EntityBoundMode.POSITION -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			return@p self.geometry.lineIntersectsRect(line, AABB.withSide(it.position, 0.0))
		}
	}

	val entities = self.entityIndex.query(line, innerPredicate)

	when (order) {
		Order.WHATEVER -> {} // as-is
		Order.RANDOM -> entities.shuffle(args.lua.random)
		Order.NEAREST -> {
			// TODO
			entities.sortWith { o1, o2 -> self.geometry.diff(o1.position, line.p0).lengthSquared.compareTo(self.geometry.diff(o2.position, line.p0).lengthSquared) }
		}
	}

	return args.lua.returnQueryResult(entities)
}

private fun queryPolyImpl(self: World<*, *>, predicate: Predicate<AbstractEntity>, args: LuaThread.ArgStack): Int {
	val poly = args.nextPoly()
	val (includeTypes, withoutEntityId, boundMode, scriptData, order) = genericQueryData(args)

	val innerPredicate: Predicate<AbstractEntity> = when (boundMode) {
		EntityBoundMode.META_BOUNDING_BOX -> Predicate p@{
			it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)
		}

		EntityBoundMode.COLLISION_AREA -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			var collisionArea = it.collisionArea

			if (collisionArea.isEmpty)
				collisionArea = it.metaBoundingBox

			return@p self.geometry.polyIntersectsPoly(poly, Poly(collisionArea))
		}

		EntityBoundMode.POSITION -> Predicate p@{
			if (!it.genericCheck(predicate, scriptData, withoutEntityId, includeTypes)) return@p false
			return@p self.geometry.polyContains(poly, it.position)
		}
	}

	val entities = self.entityIndex.query(poly.aabb, innerPredicate)

	when (order) {
		Order.WHATEVER -> {} // as-is
		Order.RANDOM -> entities.shuffle(args.lua.random)
		Order.NEAREST -> {
			val nearest = poly.centre
			entities.sortWith { o1, o2 -> self.geometry.diff(o1.position, nearest).lengthSquared.compareTo(self.geometry.diff(o2.position, nearest).lengthSquared) }
		}
	}

	return args.lua.returnQueryResult(entities)
}

private fun queryObjectAABBImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val name = args.nextOptionalString()
	return queryAABBImpl(self, Predicate { it is WorldObject && (name == null || it.config.key == name) }, args)
}

private fun queryObjectLineImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val name = args.nextOptionalString()
	return queryLineImpl(self, Predicate { it is WorldObject && (name == null || it.config.key == name) }, args)
}

private fun queryObjectRadiusImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val name = args.nextOptionalString()
	return queryRadiusImpl(self, Predicate { it is WorldObject && (name == null || it.config.key == name) }, args)
}

private fun queryObjectPolyImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val name = args.nextOptionalString()
	return queryRadiusImpl(self, Predicate { it is WorldObject && (name == null || it.config.key == name) }, args)
}

private fun queryLoungeableAABBImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val orientation = LoungeOrientation.entries[args.nextInt()]
	return queryAABBImpl(self, Predicate { it is LoungeableObject && (orientation == LoungeOrientation.NONE || it.sitOrientation == orientation) }, args)
}

private fun queryLoungeableLineImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val orientation = LoungeOrientation.entries[args.nextInt()]
	return queryLineImpl(self, Predicate { it is LoungeableObject && (orientation == LoungeOrientation.NONE || it.sitOrientation == orientation) }, args)
}

private fun queryLoungeableRadiusImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val orientation = LoungeOrientation.entries[args.nextInt()]
	return queryRadiusImpl(self, Predicate { it is LoungeableObject && (orientation == LoungeOrientation.NONE || it.sitOrientation == orientation) }, args)
}

private fun queryLoungeablePolyImpl(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val orientation = LoungeOrientation.entries[args.nextInt()]
	return queryRadiusImpl(self, Predicate { it is LoungeableObject && (orientation == LoungeOrientation.NONE || it.sitOrientation == orientation) }, args)
}

private fun objectAt(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.entityIndex.tileEntityAt(args.nextVector2i())?.entityID?.toLong())
	return 1
}

private fun entityExists(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.entities.containsKey(args.nextInt()))
	return 1
}

private fun entityCanDamage(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val a = self.entities[args.nextInt()]
	val b = self.entities[args.nextInt()]

	args.lua.push(a != null && b != null && a.team.get().canDamage(b.team.get(), a == b))
	return 1
}

private fun entityDamageTeam(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] ?: return 0
	args.lua.push(Starbound.gson.toJsonTree(entity.team.get()))
	return 1
}

private fun entityType(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] ?: return 0
	args.lua.push(entity.type.jsonName)
	return 1
}

private fun entityPosition(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] ?: return 0
	args.lua.push(entity.position)
	return 1
}

private fun entityVelocity(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? DynamicEntity ?: return 0
	args.lua.push(entity.movement.velocity)
	return 1
}

private fun entityMetaBoundBox(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] ?: return 0
	args.lua.push(entity.metaBoundingBox)
	return 1
}

private fun entityCurrency(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val currency = args.nextString()
	val entity = self.entities[id] as? PlayerEntity ?: return 0
	args.lua.push(entity.inventory.currencies[currency])
	return 1
}

private fun entityHasCountOfItem(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val exactMatch = args.nextOptionalBoolean() ?: false
	val entity = self.entities[id] as? PlayerEntity ?: return 0
	args.lua.push(entity.inventory.hasCountOfItem(desc, exactMatch))
	return 1
}

private fun entityHealth(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ActorEntity ?: return 0
	args.lua.pushTable(2)
	args.lua.setTableValue(1L, entity.health)
	args.lua.setTableValue(2L, entity.maxHealth)
	return 1
}

private fun entitySpecies(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? HumanoidActorEntity ?: return 0
	args.lua.push(entity.humanoidIdentity.species.key.left())
	return 1
}

private fun entityGender(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? HumanoidActorEntity ?: return 0
	args.lua.push(entity.humanoidIdentity.gender.jsonName)
	return 1
}

private fun entityName(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] ?: return 0

	// TODO
	when (entity) {
		is ActorEntity -> args.lua.push(entity.name)
		is WorldObject -> args.lua.push(entity.config.key)
		else -> return 0
	}

	return 1
}

private fun entityDescription(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] ?: return 0

	if (entity is InspectableEntity) {
		val species = args.nextOptionalString()
		args.lua.push(entity.inspectionDescription(species))
	} else {
		args.lua.push(entity.description)
	}

	return 1
}

private fun entityPortrait(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val mode = args.nextString()
	val entity = self.entities[id] as? ActorEntity ?: return 0

	val parts = entity.portrait(ActorEntity.PortraitMode.entries.valueOf(mode))

	args.lua.pushTable(parts.size)

	for ((i, part) in parts.withIndex()) {
		args.lua.setTableValue(i + 1L, part.toJson())
	}

	return 1
}

private fun entityHandItem(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val primary = args.nextBoolean()

	val entity = self.entities[id] as? HumanoidActorEntity ?: return 0

	if (primary) {
		args.lua.push(entity.primaryHandItem.entry.nameOrNull)
	} else {
		args.lua.push(entity.secondaryHandItem.entry.nameOrNull)
	}

	return 1
}

private fun entityHandItemDescriptor(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val primary = args.nextBoolean()

	val entity = self.entities[id] as? HumanoidActorEntity ?: return 0

	if (primary) {
		entity.primaryHandItem.createDescriptor().store(args.lua)
	} else {
		entity.secondaryHandItem.createDescriptor().store(args.lua)
	}

	return 1
}

private fun entityUniqueId(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	args.lua.push(self.entities[id]?.uniqueID?.get())
	return 1
}

private fun getObjectParameter(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val parameter = args.nextString()

	// FIXME: this is stupid (defaultValue is ignored when we lookup parameter on non existing entity),
	//  but we must retain original behavior
	val entity = self.entities[id] as? WorldObject ?: return 0
	val result = entity.lookupPropertyOrNull(parameter)

	if (result == null) {
		// TODO: while this is considerably faster, it does not correspond to original engine behavior where "default value" is always copied (lua -> json -> lua)
		if (args.top == 2) {
			args.lua.push()
		} else if (args.top != 3) {
			args.lua.dup(3)
		}
	} else {
		args.lua.push(result)
	}

	return 1
}

private fun objectSpaces(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? WorldObject

	if (entity == null) {
		args.lua.pushTable()
	} else {
		args.lua.pushTable(entity.occupySpaces.size)

		for ((i, space) in entity.occupySpaces.withIndex()) {
			args.lua.setTableValue(i + 1L, space - entity.tilePosition)
		}
	}

	return 1
}

private fun containerSize(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ContainerObject ?: return 0
	args.lua.push(entity.items.size.toLong())
	return 1
}

private fun containerClose(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ContainerObject

	if (entity == null) {
		args.lua.push(false)
		return 1
	}

	// FIXME: this doesn't get networked if called on client
	//  (AND this is the reason why in multiplayer player can't see chest/container open animations when
	//  other players open them)
	entity.closeContainer()
	args.lua.push(true)
	return 1
}

private fun containerOpen(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ContainerObject

	if (entity == null) {
		args.lua.push(false)
		return 1
	}

	// FIXME: this doesn't get networked if called on client
	//  (AND this is the reason why in multiplayer player can't see chest/container open animations when
	//  other players open them)
	entity.openContainer()
	args.lua.push(true)
	return 1
}

private fun containerItems(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ContainerObject ?: return 0

	args.lua.pushTable()
	var i = 1L

	for (item in entity.items) {
		if (item.isNotEmpty) {
			args.lua.push(i++)
			item.createDescriptor().store(args.lua)
			args.lua.setTableValue()
		}
	}

	return 1
}

private fun containerItemAt(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val slot = args.nextInt()
	val entity = self.entities[id] as? ContainerObject ?: return 0

	if (slot in 0 until entity.items.size) {
		entity.items[slot].createDescriptor().store(args.lua)
		return 1
	}

	return 0
}

private fun containerConsume(self: World<*, *>, async: Boolean, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val exactMatch = args.nextOptionalBoolean() ?: false
	val entity = self.entities[id] as? ContainerObject ?: return 0

	if (async) {
		args.lua.push(
			LuaFuture(
				future = entity.takeItem(desc, exact = exactMatch),
				isLocal = false,
				handler = {
					push(it)
					1
				}
			)
		)
	} else {
		args.lua.push(entity.takeItem(desc, exact = exactMatch).getNow(null) as Boolean?)
	}

	return 1
}

private fun containerConsumeAt(self: World<*, *>, async: Boolean, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val slot = args.nextInt()
	val amount = args.nextLong()
	val entity = self.entities[id] as? ContainerObject ?: return 0

	if (async) {
		args.lua.push(
			LuaFuture(
				future = entity.takeItemAt(slot, amount),
				isLocal = false,
				handler = {
					args.lua.push(it)
					1
				}
			)
		)
	} else {
		args.lua.push(entity.takeItemAt(slot, amount).getNow(null) as Boolean?)
	}

	return 1
}

private fun containerAvailable(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val entity = self.entities[id] as? ContainerObject ?: return 0
	args.lua.push(entity.items.take(desc, simulate = true))
	return 1
}

// why we have containerItems + containerTakeAll, when we could have containerItems + containerClear?????????
private fun containerTakeAll(self: World<*, *>, async: Boolean, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ContainerObject ?: return 0

	if (async) {
		args.lua.push(
			LuaFuture(
				future = entity.clearContainer(),
				isLocal = false,
				handler = {
					args.lua.pushTable(it.size)

					for ((i, item) in it.withIndex()) {
						args.lua.push(i + 1L)
						item.createDescriptor().store(args.lua)
						args.lua.setTableValue()
					}

					1
				}
			)
		)
	} else {
		val items = entity.clearContainer().getNow(listOf())
		args.lua.pushTable(items.size)

		for ((i, item) in items.withIndex()) {
			args.lua.push(i + 1L)
			item.createDescriptor().store(args.lua)
			args.lua.setTableValue()
		}
	}

	return 1
}

private fun containerTakeAt(self: World<*, *>, async: Boolean, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val slot = args.nextInt()
	val entity = self.entities[id] as? ContainerObject ?: return 0

	if (slot in 0 until entity.items.size) {
		if (async) {
			args.lua.push(
				LuaFuture(
					future = entity.takeItemAt(slot),
					isLocal = false,
					handler = {
						push(it)
						1
					}
				)
			)
		} else {
			args.lua.push(entity.takeItemAt(slot).getNow(null))
		}

		return 1
	}

	return 0
}

private fun containerTakeNumItemsAt(self: World<*, *>, async: Boolean, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val slot = args.nextInt()
	val amount = args.nextInt()

	require(amount >= 0L) { "Invalid amount to take: $amount" }
	val entity = self.entities[id] as? ContainerObject ?: return 0

	if (amount > 0L && slot in 0 until entity.items.size) {
		if (async) {
			args.lua.push(
				LuaFuture(
					future = entity.takeItemAt(slot, amount.toLong()),
					isLocal = false,
					handler = {
						push(it)
						1
					}
				)
			)
		} else {
			args.lua.push(entity.takeItemAt(slot, amount.toLong()).getNow(null))
		}

		return 1
	}

	return 0
}

private fun containerItemsCanFit(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val entity = self.entities[id] as? ContainerObject ?: return run { desc.store(args.lua); 1 }
	val build = desc.build(random = args.lua.random)
	val (leftover, _) = entity.items.add(build, simulate = true)

	args.lua.push(build.size - leftover.size)
	return 1
}

private fun containerItemsFitWhere(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val entity = self.entities[id] as? ContainerObject ?: return run { desc.store(args.lua); 1 }
	val build = desc.build(random = args.lua.random)
	val (leftover, touched) = entity.items.add(build, simulate = true)

	args.lua.pushTable(hashSize = 2)

	args.lua.setTableValue("leftover", leftover.size)
	args.lua.push("slots")
	args.lua.pushTable(touched.size)

	for ((i, v) in touched.withIndex()) {
		args.lua.setTableValue(i + 1L, v.toLong())
	}

	args.lua.setTableValue()

	return 1
}

private fun slotlessContainerBinding(
	self: World<*, *>,
	async: Boolean,
	dispatch: (ContainerObject, ItemStack) -> CompletableFuture<ItemStack>,
	args: LuaThread.ArgStack
): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val entity = self.entities[id] as? ContainerObject ?: return run { desc.store(args.lua); 1 }
	val build = desc.build(random = args.lua.random)

	if (async) {
		args.lua.push(
			LuaFuture(
				future = dispatch(entity, build).thenApply { it.createDescriptor() },
				isLocal = false,
				handler = {
					it.store(args.lua)
					1
				}
			)
		)

		return 1
	} else {
		dispatch(entity, build).getNow(build).createDescriptor().store(args.lua)
		return 1
	}
}

private fun slottedContainerBinding(
	self: World<*, *>,
	async: Boolean,
	dispatch: (ContainerObject, Int, ItemStack) -> CompletableFuture<ItemStack>,
	args: LuaThread.ArgStack
): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val slot = args.nextInt()
	val entity = self.entities[id] as? ContainerObject ?: return run { desc.store(args.lua); 1 }
	val build = desc.build(random = args.lua.random)

	if (async) {
		args.lua.push(
			LuaFuture(
				future = dispatch(entity, slot, build).thenApply { it.createDescriptor() },
				isLocal = false,
				handler = {
					it.store(args.lua)
					1
				}
			)
		)

		return 1
	} else {
		dispatch(entity, slot, build).getNow(build).createDescriptor().store(args.lua)
		return 1
	}
}

private fun containerSwapItems(self: World<*, *>, async: Boolean, combine: Boolean, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val desc = ItemDescriptor(args)
	val slot = args.nextInt()
	// retain original behavior where engine always decodes item descriptor and guarantees to return a copy
	val entity = self.entities[id] as? ContainerObject ?: return run { desc.store(args.lua); 1 }
	val build = desc.build(random = args.lua.random)

	if (async) {
		args.lua.push(
			LuaFuture(
				future = entity.swapItems(slot, build, tryCombine = combine).thenApply { it.createDescriptor() },
				isLocal = false,
				handler = {
					it.store(args.lua)
					1
				}
			)
		)

		return 1
	} else {
		entity.swapItems(slot, build, tryCombine = combine).getNow(build).createDescriptor().store(args.lua)
		return 1
	}
}

private fun callScriptedEntity(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val function = args.nextString()
	// TODO: uncomment to retain original behavior
	val entity = self.entities[id] ?: return 0 //?: throw IllegalStateException("Entity with ID $id does not exist")

	if (entity !is ScriptedEntity)
		throw IllegalStateException("$entity is not scripted entity")

	if (entity.isRemote)
		throw IllegalStateException("$entity is not owned by this side")

	args.lua.push(entity.callScript(function, args.copyRemaining()))
	return 1
}

private fun findUniqueEntity(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = self.findUniqueEntity(args.nextString()).thenApply { KOptional.ofNullable(it) },
			isLocal = self.isServer,
			handler = {
				it.ifPresent {
					push(it)
					return@LuaFuture 1
				}

				0
			}
		)
	)

	return 1
}

private fun findUniqueEntityAsync(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = self.findUniqueEntity(args.nextString()).thenApply { KOptional.ofNullable(it) },
			isLocal = false,
			handler = {
				it.ifPresent {
					push(it)
					return@LuaFuture 1
				}

				0
			}
		)
	)

	return 1
}

private fun sendEntityMessage(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val peek = args.peek()

	if (peek == LuaType.NUMBER) {
		val entityID = args.nextInt()
		val func = args.nextString()

		args.lua.push(
			LuaFuture(
				future = self.dispatchEntityMessage(self.connectionID, entityID, func, args.copyRemaining()),
				isLocal = Connection.connectionForEntityID(entityID) == self.connectionID,
				handler = {
					push(it)
					1
				}
			)
		)
	} else {
		val entityID = args.nextString()
		val func = args.nextString()
		val findAlreadyLoaded = self.entities.values.find { it.uniqueID.get() == entityID }

		val isLocal = if (findAlreadyLoaded == null)
			self.isServer
		else
			!findAlreadyLoaded.isRemote

		args.lua.push(
			LuaFuture(
				future = self.dispatchEntityMessage(self.connectionID, entityID, func, args.copyRemaining()),
				isLocal = isLocal,
				handler = {
					push(it)
					1
				}
			)
		)
	}

	return 1
}

private fun isMonster(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.entities[args.nextInt()] is MonsterEntity)
	return 1
}

private fun monsterType(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val entity = self.entities[args.nextInt()] as? MonsterEntity ?: return 0
	args.lua.push(entity.variant.type)
	return 1
}

private fun npcType(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val entity = self.entities[args.nextInt()] as? NPCEntity ?: return 0
	args.lua.push(entity.variant.typeName)
	return 1
}

private fun stagehandType(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val entity = self.entities[args.nextInt()] as? StagehandEntity ?: return 0
	args.lua.push(entity.typeName)
	return 1
}

private fun isNpc(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.entities[args.nextInt()] is NPCEntity)
	return 1
}

private fun isEntityInteractive(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val entity = self.entities[args.nextInt()]

	if (entity is InteractiveEntity) {
		args.lua.push(entity.isInteractive)
		return 1
	}

	return 0
}

private fun entityMouthPosition(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val entity = self.entities[args.nextInt()] ?: return 0
	// original engine returns non nil only for "Chatty entity"
	args.lua.push(entity.mouthPosition)
	return 1
}

fun provideWorldEntitiesBindings(self: World<*, *>, lua: LuaThread) {
	lua.pushBinding(self, "queryObjectAABBImpl", ::queryObjectAABBImpl)
	lua.pushBinding(self, "queryObjectLineImpl", ::queryObjectLineImpl)
	lua.pushBinding(self, "queryObjectRadiusImpl", ::queryObjectRadiusImpl)
	lua.pushBinding(self, "queryObjectPolyImpl", ::queryObjectPolyImpl)

	lua.pushBinding(self, "queryLoungeableAABBImpl", ::queryLoungeableAABBImpl)
	lua.pushBinding(self, "queryLoungeableLineImpl", ::queryLoungeableLineImpl)
	lua.pushBinding(self, "queryLoungeableRadiusImpl", ::queryLoungeableRadiusImpl)
	lua.pushBinding(self, "queryLoungeablePolyImpl", ::queryLoungeablePolyImpl)

	lua.pushBinding(self, Predicate { true }, "queryAABBImpl", ::queryAABBImpl)
	lua.pushBinding(self, Predicate { true }, "queryRadiusImpl", ::queryRadiusImpl)
	lua.pushBinding(self, Predicate { true }, "queryLineImpl", ::queryLineImpl)
	lua.pushBinding(self, Predicate { true }, "queryPolyImpl", ::queryPolyImpl)

	lua.pushBinding(self, Predicate { it is NPCEntity }, "npcQueryAABBImpl", ::queryAABBImpl)
	lua.pushBinding(self, Predicate { it is NPCEntity }, "npcQueryRadiusImpl", ::queryRadiusImpl)
	lua.pushBinding(self, Predicate { it is NPCEntity }, "npcQueryLineImpl", ::queryLineImpl)
	lua.pushBinding(self, Predicate { it is NPCEntity }, "npcQueryPolyImpl", ::queryPolyImpl)

	lua.pushBinding(self, Predicate { it is MonsterEntity }, "monsterQueryAABBImpl", ::queryAABBImpl)
	lua.pushBinding(self, Predicate { it is MonsterEntity }, "monsterQueryRadiusImpl", ::queryRadiusImpl)
	lua.pushBinding(self, Predicate { it is MonsterEntity }, "monsterQueryLineImpl", ::queryLineImpl)
	lua.pushBinding(self, Predicate { it is MonsterEntity }, "monsterQueryPolyImpl", ::queryPolyImpl)

	lua.pushBinding(self, Predicate { it is PlayerEntity }, "playerQueryAABBImpl", ::queryAABBImpl)
	lua.pushBinding(self, Predicate { it is PlayerEntity }, "playerQueryRadiusImpl", ::queryRadiusImpl)
	lua.pushBinding(self, Predicate { it is PlayerEntity }, "playerQueryLineImpl", ::queryLineImpl)
	lua.pushBinding(self, Predicate { it is PlayerEntity }, "playerQueryPolyImpl", ::queryPolyImpl)

	lua.pushBinding(self, Predicate { it is ItemDropEntity }, "itemDropQueryAABBImpl", ::queryAABBImpl)
	lua.pushBinding(self, Predicate { it is ItemDropEntity }, "itemDropQueryRadiusImpl", ::queryRadiusImpl)
	lua.pushBinding(self, Predicate { it is ItemDropEntity }, "itemDropQueryLineImpl", ::queryLineImpl)
	lua.pushBinding(self, Predicate { it is ItemDropEntity }, "itemDropQueryPolyImpl", ::queryPolyImpl)

	lua.pushBinding(self, "objectAt", ::objectAt)
	lua.pushBinding(self, "entityExists", ::entityExists)
	lua.pushBinding(self, "entityCanDamage", ::entityCanDamage)
	lua.pushBinding(self, "entityDamageTeam", ::entityDamageTeam)
	lua.pushBinding(self, "entityType", ::entityType)
	lua.pushBinding(self, "entityPosition", ::entityPosition)
	lua.pushBinding(self, "entityVelocity", ::entityVelocity)
	lua.pushBinding(self, "entityMetaBoundBox", ::entityMetaBoundBox)
	lua.pushBinding(self, "entityCurrency", ::entityCurrency)
	lua.pushBinding(self, "entityHasCountOfItem", ::entityHasCountOfItem)
	lua.pushBinding(self, "entityHealth", ::entityHealth)
	lua.pushBinding(self, "entitySpecies", ::entitySpecies)
	lua.pushBinding(self, "entityGender", ::entityGender)
	lua.pushBinding(self, "entityName", ::entityName)
	lua.pushBinding(self, "entityDescription", ::entityDescription)
	lua.pushBinding(self, "entityPortrait", ::entityPortrait)
	lua.pushBinding(self, "entityHandItem", ::entityHandItem)
	lua.pushBinding(self, "entityHandItemDescriptor", ::entityHandItemDescriptor)
	lua.pushBinding(self, "entityUniqueId", ::entityUniqueId)
	lua.pushBinding(self, "getObjectParameter", ::getObjectParameter)
	lua.pushBinding(self, "objectSpaces", ::objectSpaces)

	lua.pushBinding(self, "containerSize", ::containerSize)
	lua.pushBinding(self, "containerClose", ::containerClose)
	lua.pushBinding(self, "containerOpen", ::containerOpen)
	lua.pushBinding(self, "containerItems", ::containerItems)
	lua.pushBinding(self, "containerItemAt", ::containerItemAt)
	lua.pushBinding(self, false, "containerConsume", ::containerConsume)
	lua.pushBinding(self, true, "containerConsumeAsync", ::containerConsume)
	lua.pushBinding(self, false, "containerConsumeAt", ::containerConsumeAt)
	lua.pushBinding(self, true, "containerConsumeAtAsync", ::containerConsumeAt)
	lua.pushBinding(self, "containerAvailable", ::containerAvailable)
	lua.pushBinding(self, false, "containerTakeAll", ::containerTakeAll)
	lua.pushBinding(self, true, "containerTakeAllAsync", ::containerTakeAll)
	lua.pushBinding(self, false, "containerTakeAt", ::containerTakeAt)
	lua.pushBinding(self, true, "containerTakeAtAsync", ::containerTakeAt)
	lua.pushBinding(self, false, "containerTakeNumItemsAt", ::containerTakeNumItemsAt)
	lua.pushBinding(self, true, "containerTakeNumItemsAtAsync", ::containerTakeNumItemsAt)

	lua.pushBinding(self, "containerItemsCanFit", ::containerItemsCanFit)
	lua.pushBinding(self, "containerItemsFitWhere", ::containerItemsFitWhere)

	lua.pushBinding(self, false, ContainerObject::addItems, "containerAddItems", ::slotlessContainerBinding)
	lua.pushBinding(self, true, ContainerObject::addItems, "containerAddItemsAsync", ::slotlessContainerBinding)
	lua.pushBinding(self, false, ContainerObject::stackWithExisting, "containerStackItems", ::slotlessContainerBinding)
	lua.pushBinding(self, true, ContainerObject::stackWithExisting, "containerStackItemsAsync", ::slotlessContainerBinding)
	lua.pushBinding(self, false, ContainerObject::putItems, "containerPutItemsAt", ::slottedContainerBinding)
	lua.pushBinding(self, true, ContainerObject::putItems, "containerPutItemsAtAsync", ::slottedContainerBinding)
	lua.pushBinding(self, false, true, "containerSwapItems", ::containerSwapItems)
	lua.pushBinding(self, true, true, "containerSwapItemsAsync", ::containerSwapItems)
	lua.pushBinding(self, false, false, "containerSwapItemsNoCombine", ::containerSwapItems)
	lua.pushBinding(self, true, false, "containerSwapItemsNoCombineAsync", ::containerSwapItems)
	lua.pushBinding(self, false, ContainerObject::combineItems, "containerItemApply", ::slottedContainerBinding)
	lua.pushBinding(self, true, ContainerObject::combineItems, "containerItemApplyAsync", ::slottedContainerBinding)

	lua.pushBinding(self, "callScriptedEntity", ::callScriptedEntity)

	lua.pushBinding(self, "findUniqueEntity", ::findUniqueEntity)
	lua.pushBinding(self, "findUniqueEntityAsync", ::findUniqueEntityAsync)

	lua.pushBinding(self, "sendEntityMessage", ::sendEntityMessage)

	lua.pushBinding(self, "stagehandType", ::stagehandType)
	lua.pushBinding(self, "monsterType", ::monsterType)
	lua.pushBinding(self, "npcType", ::npcType)
	lua.pushBinding(self, "isMonster", ::isMonster)
	lua.pushBinding(self, "isNpc", ::isNpc)
	lua.pushBinding(self, "isEntityInteractive", ::isEntityInteractive)
	lua.pushBinding(self, "entityMouthPosition", ::entityMouthPosition)

	lua.setTableValueToStub("entityAggressive")
	lua.setTableValueToStub("getNpcScriptParameter")
	lua.setTableValueToStub("entityTypeName")
	lua.setTableValueToStub("loungeableOccupied")
	lua.setTableValueToStub("farmableStage")
}
