package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonNull
import com.google.gson.JsonPrimitive
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.json.JsonPath
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.getVector2i
import ru.dbotthepony.kstarbound.lua.nextColor
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.lua.setTableValue
import ru.dbotthepony.kstarbound.util.SBPattern
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import kotlin.collections.forEach
import kotlin.collections.isNotEmpty
import kotlin.collections.set
import kotlin.collections.withIndex

private fun name(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.config.key)
	return 1
}

private fun directions(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.direction.numericalValue)
	return 1
}

private fun position(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.tilePosition)
	return 1
}

private fun setInteractive(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.isInteractive = args.nextBoolean()
	return 0
}

private fun uniqueId(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.uniqueID.get())
	return 1
}

private fun setUniqueId(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.uniqueID.accept(args.nextOptionalString()?.sbIntern())
	return 0
}

private fun boundBox(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.metaBoundingBox)
	return 1
}

// original engine parity, it returns occupied spaces in local coordinates
private fun spaces(self: WorldObject, local: Boolean, args: LuaThread.ArgStack): Int {
	val spaces = self.occupySpaces
	val spos = self.tilePosition

	args.lua.pushTable(spaces.size)

	if (local) {
		for ((i, pos) in spaces.withIndex()) {
			args.lua.setTableValue(i + 1, pos - spos)
		}
	} else {
		for ((i, pos) in spaces.withIndex()) {
			args.lua.setTableValue(i + 1, pos)
		}
	}

	return 1
}

private fun setProcessingDirectives(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.animator.processingDirectives = args.nextString().sbIntern()
	return 0
}

private fun setSoundEffectEnabled(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.soundEffectEnabled = args.nextBoolean()
	return 0
}

private fun smash(self: WorldObject, args: LuaThread.ArgStack): Int {
	if (args.nextOptionalBoolean() == true) {
		self.health = 0.0
	}

	self.remove(AbstractEntity.RemovalReason.DYING)
	return 0
}

private fun level(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.lookupProperty(JsonPath("level")) { JsonPrimitive(self.world.template.threatLevel) }.asDouble)
	return 1
}

private fun toAbsolutePosition(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(args.nextVector2d() + self.position)
	return 1
}

private fun say(self: WorldObject, hasPortrait: Boolean, args: LuaThread.ArgStack): Int {
	val line = args.nextString()
	val portrait = if (hasPortrait) args.nextString() else null
	val tags = Object2ObjectArrayMap<String, String>()
	val tagsType = args.peek()

	if (tagsType == LuaType.TABLE) {
		val pairs = args.readTable(
			keyVisitor = { getString(it) ?: throw IllegalArgumentException("Tag replacement table contains non-strings as keys") },
			valueVisitor = { getString(it) ?: throw IllegalArgumentException("Tag replacement table contains non-strings as keys") },
		)

		for ((k, v) in pairs) {
			tags[k] = v
		}
	} else if (!tagsType.isNothing) {
		throw IllegalArgumentException("bad argument #2 to object.say: table expected, got $tagsType")
	} else {
		args.skip()
	}

	val sayConfig = args.nextOptionalJson() ?: JsonNull.INSTANCE

	if (line.isBlank()) {
		args.lua.push(false)
	} else if (tags.isEmpty()) {
		self.addChatMessage(line, sayConfig, portrait = portrait)
		args.lua.push(true)
	} else {
		self.addChatMessage(SBPattern.of(line).resolveOrSkip(tags::get), sayConfig, portrait = portrait)
		args.lua.push(true)
	}

	return 1
}

private fun isTouching(self: WorldObject, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.world.entities[id]

	if (entity == null) {
		args.lua.push(false)
	} else {
		args.lua.push(entity.collisionArea.intersect(self.volumeBoundingBox))
	}

	return 1
}

private fun setLightColor(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.lightSourceColor = args.nextColor()
	return 0
}

private fun getLightColor(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.lightSourceColor)
	return 1
}

private fun inputNodeCount(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.inputNodes.size.toLong())
	return 1
}

private fun outputNodeCount(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.outputNodes.size.toLong())
	return 1
}

private fun getInputNodePosition(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.inputNodes[args.nextInt()].position)
	return 1
}

private fun getOutputNodePosition(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.outputNodes[args.nextInt()].position)
	return 1
}

private fun getInputNodeLevel(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.inputNodes[args.nextInt()].state)
	return 1
}

private fun getOutputNodeLevel(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.outputNodes[args.nextInt()].state)
	return 1
}

private fun isInputNodeConnected(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.inputNodes[args.nextInt()].connections.isNotEmpty())
	return 1
}

private fun isOutputNodeConnected(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.outputNodes[args.nextInt()].connections.isNotEmpty())
	return 1
}

private fun getInputNodeIds(self: WorldObject, args: LuaThread.ArgStack): Int {
	val connections = self.inputNodes[args.nextInt()].connections

	args.lua.pushTable(connections.size)

	for (connection in connections) {
		val entity = self.world.entityIndex.tileEntityAt(connection.entityLocation, WorldObject::class)

		if (entity != null) {
			args.lua.setTableValue(entity.entityID, connection.index)
		}
	}

	return 1
}

private fun getOutputNodeIds(self: WorldObject, args: LuaThread.ArgStack): Int {
	val connections = self.outputNodes[args.nextInt()].connections

	args.lua.pushTable(connections.size)

	for (connection in connections) {
		val entity = self.world.entityIndex.tileEntityAt(connection.entityLocation, WorldObject::class)

		if (entity != null) {
			args.lua.setTableValue(entity.entityID, connection.index)
		}
	}

	return 1
}

private fun setOutputNodeLevel(self: WorldObject, args: LuaThread.ArgStack): Int {
	val index = args.nextInt()
	val state = args.nextBoolean()
	self.outputNodes[index].state = state
	return 0
}

private fun setAllOutputNodes(self: WorldObject, args: LuaThread.ArgStack): Int {
	val state = args.nextBoolean()
	self.outputNodes.forEach { it.state = state }
	return 0
}

private fun setConfigParameter(self: WorldObject, args: LuaThread.ArgStack): Int {
	val key = args.nextString().sbIntern()
	val value = args.nextJson()
	self.parameters[key] = value
	return 0
}

private fun setAnimationParameter(self: WorldObject, args: LuaThread.ArgStack): Int {
	val key = args.nextString().sbIntern()
	val value = args.nextJson()
	self.scriptedAnimationParameters[key] = value
	return 0
}

private fun setMaterialSpaces(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.declaredMaterialSpaces.clear()

	if (args.hasNext) {
		val pairs = args.readTableValues { tI ->
			push(1L)
			val vType = loadTableValue(tI)
			val position = getVector2i() ?: throw IllegalArgumentException("invalid space pair, first value is not a vector: $vType")

			push(2L)
			val mType = loadTableValue(tI)
			val material = getString() ?: throw IllegalArgumentException("invalid space pair, second value is not a string: $mType")

			position to Registries.tiles.ref(material)
		}

		self.declaredMaterialSpaces.addAll(pairs)
	}

	return 0
}

private fun setDamageSources(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.customDamageSources.clear()

	if (args.hasNext) {
		val sources = args.readTableValues {
			Starbound.gson.fromJsonFast(getJson(it)!!, DamageSource::class.java)
		}

		self.customDamageSources.addAll(sources)
	}

	return 0
}

private fun health(self: WorldObject, args: LuaThread.ArgStack): Int {
	args.lua.push(self.health)
	return 1
}

private fun setHealth(self: WorldObject, args: LuaThread.ArgStack): Int {
	self.health = args.nextDouble()
	return 0
}

fun provideWorldObjectBindings(self: WorldObject, lua: LuaThread) {
	provideConfigBinding(lua, self::lookupPropertyOrNull)

	lua.pushTable()
	lua.dup()
	lua.storeGlobal("object")

	lua.pushBinding(self, "name", ::name)
	lua.pushBinding(self, "directions", ::directions)
	lua.pushBinding(self, "position", ::position)
	lua.pushBinding(self, "setInteractive", ::setInteractive)
	lua.pushBinding(self, "uniqueId", ::uniqueId)
	lua.pushBinding(self, "setUniqueId", ::setUniqueId)
	lua.pushBinding(self, "boundBox", ::boundBox)
	lua.pushBinding(self, true, "spaces", ::spaces)
	lua.pushBinding(self, false, "worldSpaces", ::spaces)
	lua.pushBinding(self, "setProcessingDirectives", ::setProcessingDirectives)
	lua.pushBinding(self, "setSoundEffectEnabled", ::setSoundEffectEnabled)
	lua.pushBinding(self, "smash", ::smash)
	lua.pushBinding(self, "level", ::level)
	lua.pushBinding(self, "toAbsolutePosition", ::toAbsolutePosition)
	lua.pushBinding(self, false, "say", ::say)
	lua.pushBinding(self, true, "sayPortrait", ::say)
	lua.pushBinding(self, "isTouching", ::isTouching)
	lua.pushBinding(self, "setLightColor", ::setLightColor)
	lua.pushBinding(self, "getLightColor", ::getLightColor)
	lua.pushBinding(self, "inputNodeCount", ::inputNodeCount)
	lua.pushBinding(self, "outputNodeCount", ::outputNodeCount)
	lua.pushBinding(self, "getInputNodePosition", ::getInputNodePosition)
	lua.pushBinding(self, "getOutputNodePosition", ::getOutputNodePosition)
	lua.pushBinding(self, "getInputNodeLevel", ::getInputNodeLevel)
	lua.pushBinding(self, "getOutputNodeLevel", ::getOutputNodeLevel)
	lua.pushBinding(self, "isInputNodeConnected", ::isInputNodeConnected)
	lua.pushBinding(self, "isOutputNodeConnected", ::isOutputNodeConnected)
	lua.pushBinding(self, "getInputNodeIds", ::getInputNodeIds)
	lua.pushBinding(self, "getOutputNodeIds", ::getOutputNodeIds)
	lua.pushBinding(self, "setOutputNodeLevel", ::setOutputNodeLevel)
	lua.pushBinding(self, "setAllOutputNodes", ::setAllOutputNodes)
	lua.pushBinding(self, WorldObject::offeredQuests, "setOfferedQuests", ::setOfferedQuests)
	lua.pushBinding(self, WorldObject::turnInQuests, "setTurnInQuests", ::setTurnInQuests)
	lua.pushBinding(self, "setConfigParameter", ::setConfigParameter)
	lua.pushBinding(self, "setAnimationParameter", ::setAnimationParameter)
	lua.pushBinding(self, "setMaterialSpaces", ::setMaterialSpaces)
	lua.pushBinding(self, "setDamageSources", ::setDamageSources)
	lua.pushBinding(self, "health", ::health)
	lua.pushBinding(self, "setHealth", ::setHealth)

	lua.pop()
}
