package ru.dbotthepony.kstarbound.world.entities

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import org.classdump.luna.ByteString
import org.classdump.luna.Table
import ru.dbotthepony.kommons.collect.filterNotNull
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.defs.DamageNotification
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.defs.EntityDamageTeam
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.defs.HitType
import ru.dbotthepony.kstarbound.defs.InteractAction
import ru.dbotthepony.kstarbound.defs.InteractRequest
import ru.dbotthepony.kstarbound.defs.actor.HumanoidConfig
import ru.dbotthepony.kstarbound.defs.actor.HumanoidEmote
import ru.dbotthepony.kstarbound.defs.actor.HumanoidIdentity
import ru.dbotthepony.kstarbound.defs.actor.NPCVariant
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.defs.quest.QuestArcDescriptor
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.io.BinaryStringCodec
import ru.dbotthepony.kstarbound.io.NullableBinaryStringCodec
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.LuaMessageHandlerComponent
import ru.dbotthepony.kstarbound.lua.LuaUpdateComponent
import ru.dbotthepony.kstarbound.lua.bindings.MovementControllerBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideConfigBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideEntityBindings
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.tableMapOf
import ru.dbotthepony.kstarbound.lua.tableOf
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.network.packets.DamageRequestPacket
import ru.dbotthepony.kstarbound.network.syncher.EventCounterElement
import ru.dbotthepony.kstarbound.network.syncher.FloatingNetworkedElement
import ru.dbotthepony.kstarbound.network.syncher.NetworkedList
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedJsonElement
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import ru.dbotthepony.kstarbound.util.GameTimer
import ru.dbotthepony.kstarbound.util.random.staticRandomInt
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.api.InteractiveEntity
import ru.dbotthepony.kstarbound.world.entities.api.ScriptedEntity
import java.io.DataOutputStream
import kotlin.math.absoluteValue

class NPCEntity(val variant: NPCVariant) : HumanoidActorEntity() {
	override val type: EntityType
		get() = EntityType.NPC
	override val damageBarType: DamageBarType
		get() = DamageBarType.DEFAULT

	override fun writeNetwork(stream: DataOutputStream, isLegacy: Boolean) {
		variant.write(stream, isLegacy)
	}

	override val metaBoundingBox: AABB
		get() = AABB.withSide(position, 4.0, 4.0)
	override val name: String
		get() = variant.humanoidIdentity.name
	override val description: String
		get() = "Some funny looking person" // FIXME: this is how it is in original engine

	override val humanoidIdentity: HumanoidIdentity
		get() = variant.humanoidIdentity
	override val humanoidConfig: HumanoidConfig
		get() = variant.humanoidConfig

	override val blinkInterval: Vector2d
		get() = Globals.npcs.blinkInterval

	override var xAimPosition by networkGroup.upstream.add(FloatingNetworkedElement.fixed(0.0625).also { it.interpolator = Interpolator.Linear })
	override var yAimPosition by networkGroup.upstream.add(FloatingNetworkedElement.fixed(0.0625).also { it.interpolator = Interpolator.Linear })

	init {
		networkGroup.upstream.add(uniqueID)
		networkGroup.upstream.add(team)
		humanoidNetStates.forEach { networkGroup.upstream.add(it) }
	}

	val newChatMessageEvent = networkGroup.upstream.add(EventCounterElement())
	var chatMessage by networkGroup.upstream.add(networkedString())
	var chatPortrait by networkGroup.upstream.add(networkedString())
	var chatConfig by networkGroup.upstream.add(networkedJsonElement())

	var statusText by networkGroup.upstream.add(networkedData(null, NullableBinaryStringCodec))
	var displayNametag by networkGroup.upstream.add(networkedBoolean())
	override var isInteractive by networkGroup.upstream.add(networkedBoolean())

	val offeredQuests = networkGroup.upstream.add(NetworkedList(QuestArcDescriptor.CODEC, QuestArcDescriptor.LEGACY_CODEC))
	val turnInQuests = networkGroup.upstream.add(NetworkedList(BinaryStringCodec))

	var isShifting by networkGroup.upstream.add(networkedBoolean())
	var damageOnTouch by networkGroup.upstream.add(networkedBoolean())
	override var disableWornArmor by networkGroup.upstream.add(networkedBoolean(variant.disableWornArmor))
	var deathParticleBurst by networkGroup.upstream.add(networkedData(variant.humanoidConfig.deathParticles, NullableBinaryStringCodec))

	val dropPools = networkGroup.upstream.add(NetworkedList(Registries.treasurePools.nameRefCodec))
	var isAggressive by networkGroup.upstream.add(networkedBoolean())

	override val statusController = StatusController(this, variant.statusControllerSettings)

	override val emoteCooldownTimer = GameTimer(Globals.npcs.emoteCooldown)
	override val danceTimer = GameTimer()

	init {
		networkGroup.upstream.add(movement.networkGroup)
		networkGroup.upstream.add(effects.networkGroup)
		networkGroup.upstream.add(statusController)
		networkGroup.upstream.add(armorNetworkGroup)
		networkGroup.upstream.add(toolsNetworkGroup)

		dropPools.addAll(variant.dropPools)
		team.accept(variant.team)

		movement.applyParameters(variant.movementParameters)

		if (variant.movementParameters.physicsEffectCategories == null) {
			movement.applyParameters(ActorMovementParameters(physicsEffectCategories = ImmutableSet.of("npc")))
		}

		statusController.addPersistentEffects("innate", variant.innateStatusEffects)
		statusController.addPersistentEffects("species", variant.species.value.buildStatusEffects())
		statusController.setProperty("species", JsonPrimitive(variant.species.key))

		if (!statusController.hasProperty("effectDirectives"))
			statusController.setProperty("effectDirectives", JsonPrimitive(variant.species.value.effectDirectives))
	}

	private var hitDamageNotificationLimiter = 0
	override var isPersistent: Boolean = variant.persistent
	override var keepAlive: Boolean = variant.keepAlive

	override val lua = LuaEnvironment()
	val luaUpdate = LuaUpdateComponent(lua)
	val luaMovement = MovementControllerBindings(movement)
	val luaMessages = LuaMessageHandlerComponent(lua) { toString() }

	init {
		lua.globals["storage"] = lua.tableOf()
	}

	override fun handleMessage(connection: Int, message: String, arguments: JsonArray): JsonElement? {
		return luaMessages.handle(message, connection == connectionID, arguments) ?: statusController.handleMessage(message, connection == connectionID, arguments)
	}

	override fun move(delta: Double) {
		luaMovement.apply()

		// This differs from original engine because it may allow 1 tick long conflicting
		// lounging even for local entities because they are updated in parallel
		// Shouldn't cause issues considering such situation is also possible on original engine
		// when seat is contested in multiplayer by two players
		if (isLocal && inConflictingLoungeAnchor())
			movement.anchorNetworkState = null

		super.move(delta)
	}

	@JsonFactory
	data class SerializedData(
		// val npcVariant: NPCVariant,
		val movementController: ActorMovementController.SerializedData,
		val statusController: StatusController.SerializedData,
		val aimPosition: Vector2d = Vector2d.ZERO,
		val humanoidState: HumanoidState = HumanoidState.IDLE,
		val humanoidEmoteState: HumanoidEmote = HumanoidEmote.IDLE,
		val isInteractive: Boolean = false,
		val shifting: Boolean = false,
		val damageOnTouch: Boolean = false,
		val effectEmitter: JsonElement,
		val armor: ArmorSerializedData = ArmorSerializedData(),
		val tools: ToolSerializedData = ToolSerializedData(),
		val disableWornArmor: Boolean,
		val uniqueId: String? = null,
		val team: EntityDamageTeam? = null,
		val deathParticleBurst: String? = null,
		val dropPools: ImmutableList<Registry.Ref<TreasurePoolDefinition>> = ImmutableList.of(),
		val aggressive: Boolean = false,
	)

	override fun deserialize(data: JsonObject) {
		super.deserialize(data)

		val serialized = Starbound.gson.fromJsonFast(data, SerializedData::class.java)

		movement.deserialize(serialized.movementController)
		statusController.deserialize(serialized.statusController)
		aimPosition = serialized.aimPosition
		humanoidState = serialized.humanoidState
		humanoidEmote = serialized.humanoidEmoteState
		isInteractive = serialized.isInteractive
		isShifting = serialized.shifting
		damageOnTouch = serialized.damageOnTouch

		deserializeArmorUser(serialized.armor)
		deserializeToolUser(serialized.tools)

		disableWornArmor = serialized.disableWornArmor
		uniqueID.accept(serialized.uniqueId)

		if (serialized.team != null)
			team.accept(serialized.team)

		deathParticleBurst = serialized.deathParticleBurst
		dropPools.clear()
		dropPools.addAll(serialized.dropPools.filter { it.isPresent })
		isAggressive = serialized.aggressive
		blinkTimer.reset(0.0)
	}

	override fun serialize(data: JsonObject) {
		super.serialize(data)

		// Very ugly
		data["npcVariant"] = Starbound.gson.toJsonTree(variant)

		val serialized = SerializedData(
			movementController = movement.serialize(),
			statusController = statusController.serialize(),
			aimPosition = aimPosition,
			humanoidState = humanoidState,
			humanoidEmoteState = humanoidEmote,
			isInteractive = isInteractive,
			shifting = isShifting,
			damageOnTouch = damageOnTouch,
			effectEmitter = effects.serialize(),
			armor = serializeArmorUser(),
			tools = serializeToolUser(),
			disableWornArmor = disableWornArmor,
			uniqueId = uniqueID.get(),
			team = team.get(),
			deathParticleBurst = deathParticleBurst,
			dropPools = dropPools.stream().filter { it.isPresent }.collect(ImmutableList.toImmutableList()),
			aggressive = isAggressive,
		)

		mergeJson(data, Starbound.gson.toJsonTree(serialized))
	}

	override fun onJoinWorld(world: World<*, *>) {
		super.onJoinWorld(world)

		if (!isRemote) {
			for ((slot, item) in variant.items) {
				setItem(slot, item)
			}

			provideEntityBindings(this, lua)

			provideConfigBindings(lua) { key, default ->
				key.find(variant.scriptConfig) ?: default
			}

			BehaviorState.provideBindings(lua)

			luaUpdate.stepCount = variant.initialScriptDelta
			lua.attach(variant.scripts)

			luaMovement.init(lua)
			lua.init()
		}
	}

	fun addChatMessage(message: String, config: JsonElement, portrait: String = "") {
		chatMessage = message
		chatPortrait = portrait
		chatConfig = config
		newChatMessageEvent.trigger()
	}

	override fun potentiallyCanBeHit(
		source: DamageSource,
		attacker: AbstractEntity?,
		inflictor: AbstractEntity?
	): Boolean {
		return !isInvulnerable && !isDead
	}

	override fun queryHit(source: DamageSource, attacker: AbstractEntity?, inflictor: AbstractEntity?): HitType? {
		if (isInvulnerable || isDead)
			return null

		if (queryShieldHit(source))
			return HitType.SHIELD_HIT

		if (source.intersect(world.geometry, movement.computeGlobalHitboxes()))
			return HitType.HIT

		return null
	}

	override fun experienceDamage(damage: DamageRequestPacket): List<DamageNotification> {
		val notifications = statusController.experienceDamage(damage.request)
		val totalDamage = notifications.sumOf { it.healthLost }

		if (totalDamage > 0.0 && hitDamageNotificationLimiter++ < Globals.npcs.hitDamageNotificationLimit) {
			lua.invokeGlobal("damage", lua.tableMapOf(
				"sourceId" to damage.request.sourceEntityId,
				"damage" to totalDamage,
				"sourceDamage" to damage.request.damage,
				"sourceKind" to damage.request.damageSourceKind
			))
		}

		return notifications
	}

	override fun tickParallel(delta: Double) {
		super.tickParallel(delta)

		if (hitDamageNotificationLimiter > 0)
			hitDamageNotificationLimiter--

		// HERE: movementSuppressed/runningSuppressed by FireableItem
		//  Implementing these skipped because they don't actually apply these suppressions to movement controller
		//  because this logic was moved to Lua scripts

		if (isLocal) {
			val anchor = movement.anchorState

			if (anchor is LoungeAnchorState) {
				if (anchor.emote != null) {
					requestEmote(anchor.emote)
				}

				humanoidState = anchor.orientation.humanoidState
				statusController.setPersistentEffects("lounging", anchor.statusEffects)
				effects.addEffectSources("normal", anchor.effectEmitters)
			} else {
				statusController.removePersistentEffects("lounging")

				if (movement.isGroundMovement) {
					if (movement.isRunning)
						humanoidState = HumanoidState.RUN
					else if (movement.isWalking)
						humanoidState = HumanoidState.WALK
					else if (movement.isCrouching)
						humanoidState = HumanoidState.DUCK
					else
						humanoidState = HumanoidState.IDLE
				} else if (movement.isLiquidMovement) {
					if (movement.velocity.dot(movement.determineGravity().unitVector).absoluteValue != 1.0)
						humanoidState = HumanoidState.SWIM
					else
						humanoidState = HumanoidState.SWIM_IDLE
				} else {
					if (movement.velocity.dot(movement.determineGravity().unitVector).absoluteValue > 0.01)
						humanoidState = HumanoidState.JUMP
					else
						humanoidState = HumanoidState.FALL
				}
			}

			if (danceTimer.tick(delta))
				humanoidDance = null

			effects.tick(delta)
		}

		// TODO: HERE: setupHumanoidClothingDrawables
	}

	override fun tick(delta: Double) {
		super.tick(delta)

		if (!isRemote) {
			if (isDead || lua.errorState) {
				remove(RemovalReason.DYING)
				return
			} else {
				val shouldDie = lua.invokeGlobal("shouldDie")

				if (shouldDie.isNotEmpty() && shouldDie[0] is Boolean && shouldDie[0] as Boolean) {
					remove(RemovalReason.DYING)
					return
				}
			}

			// luaUpdate.update(delta)
		}

		tickTools(delta, setOf())
	}

	override fun onRemove(world: World<*, *>, reason: RemovalReason) {
		super.onRemove(world, reason)

		lua.invokeGlobal("die")

		val dropPools by lazy { dropPools.stream().map { it.entry }.filterNotNull().toList() }

		if (!isRemote && reason.dying && dropPools.isNotEmpty()) {
			// TODO: what.
			val pool = dropPools[staticRandomInt(0, dropPools.size - 1, variant.seed)]

			for (item in pool.value.evaluate(world.random, variant.level)) {
				val entity = ItemDropEntity(item)
				entity.position = position
				entity.movement.xVelocity += world.random.nextDouble(-10.0, 10.0)
				entity.movement.yVelocity += world.random.nextDouble(-10.0, 10.0)
				entity.joinWorld(world)
			}
		}
	}


}
