package ru.dbotthepony.kstarbound.world.entities.player

import com.google.gson.JsonObject
import kotlinx.coroutines.runBlocking
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.defs.DamageNotification
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.defs.HitType
import ru.dbotthepony.kstarbound.defs.actor.HumanoidConfig
import ru.dbotthepony.kstarbound.defs.actor.HumanoidIdentity
import ru.dbotthepony.kstarbound.defs.actor.HumanoidEmote
import ru.dbotthepony.kstarbound.defs.actor.player.PlayerGamemode
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.network.packets.DamageRequestPacket
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedEnum
import ru.dbotthepony.kstarbound.network.syncher.networkedEnumExtraStupid
import ru.dbotthepony.kstarbound.network.syncher.networkedEventCounter
import ru.dbotthepony.kstarbound.network.syncher.networkedFixedPoint
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import ru.dbotthepony.kstarbound.util.GameTimer
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.Animator
import ru.dbotthepony.kstarbound.world.entities.HumanoidActorEntity
import ru.dbotthepony.kstarbound.world.entities.StatusController
import ru.dbotthepony.kstarbound.world.physics.Poly
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.*
import kotlin.properties.Delegates

class PlayerEntity() : HumanoidActorEntity() {
	enum class State(override val jsonName: String) : IStringSerializable {
		IDLE("Idle"),
		WALK("Walk"),
		RUN("Run"),
		JUMP("Jump"),
		FALL("Fall"),
		SWIM("Swim"),
		SWIM_IDLE("SwimIdle"),
		TELEPORT_IN("TeleportIn"),
		TELEPORT_OUT("TeleportOut"),
		CROUCH("Crouch"),
		LOUNGE("Lounge");
	}

	constructor(data: DataInputStream, isLegacy: Boolean) : this() {
		uniqueID.accept(data.readInternedString())
		description = data.readInternedString()
		gamemode = PlayerGamemode.entries[if (isLegacy) data.readInt() else data.readUnsignedByte()]
		humanoidIdentity = HumanoidIdentity.read(data, isLegacy)
		humanoidConfig = runBlocking { humanoidIdentity.species.value!!.humanoidConfig() }
	}

	constructor(data: JsonObject) : this() {
		deserialize(data)
	}

	override val type: EntityType
		get() = EntityType.PLAYER

	var gamemode = PlayerGamemode.CASUAL

	override var description: String = "This guy has nothing to say to himself."

	override fun writeNetwork(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeBinaryString(uniqueID.get() ?: "")
		stream.writeBinaryString(description)
		if (isLegacy) stream.writeInt(gamemode.ordinal) else stream.writeByte(gamemode.ordinal)
		humanoidIdentity.write(stream, isLegacy)
	}

	val inventory = PlayerInventory()
	val songbook = Songbook(this)
	val effectAnimator = if (Globals.player.effectsAnimator.value.get() == null) Animator() else Animator(Globals.player.effectsAnimator.value.get()!!)
	override val statusController = StatusController(this, Globals.player.statusControllerSettings)
	val techController = TechController(this)

	var playerState by networkGroup.upstream.add(networkedEnum(State.IDLE))
	var shifting by networkGroup.upstream.add(networkedBoolean())
	override var xAimPosition by networkGroup.upstream.add(networkedFixedPoint(0.003125))
	override var yAimPosition by networkGroup.upstream.add(networkedFixedPoint(0.003125).also { it.interpolator = Interpolator.Linear })
	override var humanoidIdentity: HumanoidIdentity by networkGroup.upstream.add(networkedData(HumanoidIdentity(), HumanoidIdentity.CODEC, HumanoidIdentity.LEGACY_CODEC))

	init {
		networkGroup.upstream.add(team)
	}

	val landed = networkGroup.upstream.add(networkedEventCounter())
	var chatMessage by networkGroup.upstream.add(networkedString())
	val newChatMessage = networkGroup.upstream.add(networkedEventCounter())
	var emote by networkGroup.upstream.add(networkedEnumExtraStupid(HumanoidEmote.IDLE))

	override val lua: LuaEnvironment
		get() = TODO("Not yet implemented")
	override val emoteCooldownTimer: GameTimer = GameTimer()
	override val danceTimer: GameTimer?
		get() = null

	override val blinkInterval: Vector2d
		get() = Globals.player.blinkInterval

	init {
		networkGroup.upstream.add(inventory.networkGroup)
		networkGroup.upstream.add(toolsNetworkGroup)
		networkGroup.upstream.add(armorNetworkGroup)
		networkGroup.upstream.add(songbook.networkGroup)
		networkGroup.upstream.add(movement.networkGroup)
		networkGroup.upstream.add(effects.networkGroup)
		networkGroup.upstream.add(effectAnimator.networkGroup)
		networkGroup.upstream.add(statusController)
		networkGroup.upstream.add(techController.networkGroup)

		movement.resetBaseParameters(Globals.player.movementParameters)
	}

	override var humanoidConfig: HumanoidConfig by Delegates.notNull()
		private set
	override val damageBarType: DamageBarType
		get() = DamageBarType.DEFAULT
	override val name: String
		get() = humanoidIdentity.name

	override val metaBoundingBox: AABB
		get() = Globals.player.metaBoundBox + position

	override val isPersistent: Boolean
		get() = false

	var isAdmin = false

	val isTeleporting: Boolean
		get() = playerState == State.TELEPORT_IN || playerState == State.TELEPORT_OUT

	override val damageHitbox: List<Poly>
		get() = movement.computeGlobalHitboxes()

	override fun potentiallyCanBeHit(
		source: DamageSource,
		attacker: AbstractEntity?,
		inflictor: AbstractEntity?
	): Boolean {
		return !isAdmin && !isDead && !isTeleporting && !statusController.statPositive("invulnerable")
	}

	override fun queryHit(source: DamageSource, attacker: AbstractEntity?, inflictor: AbstractEntity?): HitType? {
		if (source.intersect(world.geometry, movement.computeGlobalHitboxes()))
			return HitType.HIT

		if (queryShieldHit(source))
			return HitType.SHIELD_HIT

		return null
	}

	override fun experienceDamage(damage: DamageRequestPacket): List<DamageNotification> {
		if (!isAdmin && !isDead && !isTeleporting && !statusController.statPositive("invulnerable"))
			return statusController.experienceDamage(damage.request)

		return listOf()
	}

	var uuid: UUID by Delegates.notNull()
		private set
}
