
-- why not use _ENV anyway lol
self = self or {}

LUA_HINT_NONE = 0
LUA_HINT_ARRAY = 1
LUA_HINT_OBJECT = 2

local rawset = rawset
local type = type
local pairs = pairs
local ipairs = ipairs
local setmetatable = setmetatable
local select = select
local math = math
local string = string
local format = string.format

function checkarg(value, index, expected, fnName, overrideExpected)
	if type(value) ~= expected then
		error(format('bad argument #%d to %s: %s expected, got %s', index, fnName, overrideExpected or expected, type(value)), 3)
	end
end

local checkarg = checkarg

-- this replicates original engine code, but it shouldn't work in first place
local function __newindex(self, key, value)
	local nils = getmetatable(self).__nils

	-- If we are setting an entry to nil, need to add a bogus integer entry
	-- to the __nils table, otherwise need to set the entry *in* the __nils
	-- table to nil and remove it.

	-- TODO: __newindex is called only when assigning non-existing keys to values,
	-- TODO: as per Lua manual.
	-- TODO: Chucklefish weren't aware of this?

	if key == nil then
		nils[key] = true
	else
		nils[key] = nil
	end

	rawset(self, key, value)
end

function jobject()
	return setmetatable({}, {
		__newindex = __newindex,
		__nils = {},
		__typehint = LUA_HINT_OBJECT
	})
end

function jarray()
	return setmetatable({}, {
		__newindex = __newindex,
		__nils = {},
		__typehint = LUA_HINT_ARRAY
	})
end

function jremove(self, key)
	checkarg(self, 1, 'table', 'jremove')
	local meta = getmetatable(self)

	if meta and meta.__nils then
		meta.__nils[key] = true
	end

	self[key] = nil
end

function jsize(self)
	checkarg(self, 1, 'table', 'jsize')

	local elemCount = 0
	local highestIndex = 0
	local hintList = false

	local meta = getmetatable(self)

	if meta then
		if meta.__typehint == LUA_HINT_ARRAY then
			hintList = true
		end

		local nils = meta.__nils

		if nils ~= nil then
			for k, v in pairs(nils) do
				if type(k) == 'number' then
					highestIndex = math.max(k, highestIndex)
				else
					hintList = false
				end

				elemCount = elemCount + 1
			end
		end
	end

	for k, v in pairs(self) do
		if type(k) == 'number' then
			highestIndex = math.max(k, highestIndex)
		else
			hintList = false
		end

		elemCount = elemCount + 1
	end

	if hintList then
		return highestIndex
	else
		return elemCount
	end
end

function jresize(self, target)
	checkarg(self, 1, 'table', 'jresize')
	checkarg(target, 2, 'number', 'jresize')

	local meta = getmetatable(self)

	if meta and meta.__nils then
		local indices = {}
		local i = 1

		for k, v in pairs(meta.__nils) do
			if type(k) == 'number' and k % 1.0 == 0.0 and k > target then
				indices[i] = k
				i = i + 1
			end
		end

		for i, v in ipars(indices) do
			meta.__nils[v] = nil
		end
	end

	local indices = {}
	local i = 1

	for k, v in pairs(self) do
		if type(k) == 'number' and k % 1.0 == 0.0 and k > target then
			indices[i] = k
			i = i + 1
		end
	end

	for i, v in ipars(indices) do
		self[v] = nil
	end
end

do
	local __print = __print
	local __print_warn = __print_warn
	local __print_error = __print_error
	local __print_fatal = __print_fatal

	function print(...)
		local values = {}

		-- this is required for properly printing nils
		for i = 1, select('#', ...) do
			values[i] = tostring(select(i, ...))
		end

		__print(table.concat(values, '\t'))
	end

	function sb.logInfo(text, ...)
		__print(format(text, ...))
	end

	function sb.logWarn(text, ...)
		__print_warn(format(text, ...))
	end

	function sb.logError(text, ...)
		__print_error(format(text, ...))
	end

	function sb.logFatal(text, ...)
		__print_fatal(format(text, ...))
	end
end

do
	local __require = __require
	local loadedScripts = {}

	function require(path, ...)
		checkarg(path, 1, 'string', 'require')

		if string.sub(path, 1, 1) ~= '/' then
			error('require: script path must be absolute: ' .. path)
		end

		if loadedScripts[path] then return unpack(loadedScripts[path]) end
		local fn = __require(path)

		if fn then
			local result = {fn(...)}
			loadedScripts[path] = result
			return unpack(result)
		else
			print('Failed to require Lua script ' .. path)
			loadedScripts[path] = {}
		end
	end
end

do
	local __random_double = __random_double
	local __random_long = __random_long
	local __random_seed = __random_seed
	local floor = math.floor

	function math.random(a, b)
		local tA = type(a)
		local tB = type(b)

		if tA == 'nil' and tB == 'nil' then
			return __random_double()
		else
			local low, high

			if tB == 'nil' then
				if tA ~= 'number' then
					error('bad argument #1 to math.random: integer expected, got ' .. tA, 2)
				end

				low = 1
				high = tA

				if high % 1.0 ~= 0.0 then
					error('bad argument #1 to math.random: integer expected, got double', 2)
				end
			else
				if tA ~= 'number' then
					error('bad argument #1 to math.random: number expected, got ' .. tA, 2)
				end

				if tA ~= 'number' then
					error('bad argument #2 to math.random: number expected, got ' .. tB, 2)
				end

				low = tA
				high = tB

				if low % 1.0 ~= 0.0 then
					error('bad argument #1 to math.random: integer expected, got double', 2)
				end

				if high % 1.0 ~= 0.0 then
					error('bad argument #2 to math.random: integer expected, got double', 2)
				end
			end

			if low > high then
				error('interval is empty (low: ' .. low .. ', high: ' .. high .. ')', 2)
			elseif low == high then
				return floor(low)
			elseif high >= 9223372036854775807 or low <= -9223372036854775808 then
				error('interval too large', 2)
			else
				return __random_long(floor(low), floor(high))
			end
		end
	end

	function math.randomseed(seed)
		if type(seed) ~= 'number' then
			error('bad argument #1 to math.randomseed: integer expected, got ' .. type(seed), 2)
		end

		if seed % 1.0 ~= 0.0 then
			error('bad argument #1 to math.randomseed: integer expected, got double', 2)
		end

		__random_seed(floor(seed))
	end
end


do
	local lerp = math.lerp
	local PI = math.pi
	local sin = math.sin

	local function lerpSin(t, a, b)
		return lerp((sin(t * PI - PI / 2.0) + 1.0) / 2.0, a, b)
	end

	function sb.interpolateSinEase(t, a, b)
		checkarg(t, 1, 'number', 'sb.interpolateSinEase')

		local tA = type(a)
		local tB = type(b)

		if tA == 'number' and tB == 'number' then
			return lerpSin(t, a, b)
		elseif tA == 'table' and tB == 'table' then
			checkarg(a[1], 2, 'number', 'sb.interpolateSinEase', 'number or vector')
			checkarg(a[2], 2, 'number', 'sb.interpolateSinEase', 'number or vector')

			checkarg(b[1], 3, 'number', 'sb.interpolateSinEase', 'number or vector')
			checkarg(b[2], 3, 'number', 'sb.interpolateSinEase', 'number or vector')

			return {lerpSin(t, a[1], b[1]), lerpSin(t, a[2], b[2])}
		else
			error('illegal arguments to sb.interpolateSinEase: ' .. tA .. ' ' .. tB)
		end
	end
end

-- why is this even a thing.
sb.print = tostring

-- KStarbound global Lua functions and extensions
do
	local min = math.min
	local max = math.max

	function math.clamp(value, _min, _max)
		if _min > _max then
			error(format('interval is empty: %d <= %d', _min, _max), 2)
		end

		return max(min(value, _min), _max)
	end
end

function math.lerp(t, a, b)
	return t * (b - a) + a
end

function table.copy(source)
	local copy = {}

	for k, v in pairs(source) do
		if type(v) == 'table' then
			copy[k] = table.copy(v)
		else
			copy[k] = v
		end
	end

	return copy
end

function jsonType(value)
	if type(value) ~= 'table' then
		return LUA_HINT_NONE
	else
		local meta = getmetatable(value)

		if not meta then
			return LUA_HINT_NONE
		else
			return meta.__typehint or LUA_HINT_NONE
		end
	end
end

function mergeJson(base, with)
	local bType = jsonType(base)
	local wType = jsonType(with)

	if bType == LUA_HINT_OBJECT and wType == LUA_HINT_OBJECT or bType == LUA_HINT_NONE and wType == LUA_HINT_NONE and type(base) == 'table' and type(with) == 'table' then
		-- merge as json objects as long as they are both "json objects" or plain tables
		for k, v in pairs(with) do
			base[k] = mergeJson(base[k], v)
		end

		return base
	elseif wType == nil then
		return base
	else
		return with
	end
end

string.__index = string

do
	local sub = string.sub

	function string:__index(key)
		if type(key) == 'number' then
			return sub(self, key, key)
		else
			return string[key]
		end
	end
end

setmetatable('', string)


